"""AxNum field type functions - before, after / insert, update, delete"""
import json
import backend.exec as ax_exec


async def before_display(db_session, field, form, current_user):
    """
    Executes python code from AxField.private_options

    Returns:
        Object: Returns updated value of current field"""
    del db_session
    if not field.private_options_json or field.private_options_json == '{}':
        return None

    try:
        options = json.loads(field.private_options_json)
        code = options['code']
        ax = await ax_exec.execute_field_code(
            code=code,
            form=form,
            current_user=current_user)
        field.value = ax.value
        return field.value
    except Exception as exc:
        return (f"There was error with executing AxHtml."
                f"Please check field options: {exc}")



""" Local fork of sanic_graphql lib """

import re
import time
from collections import Mapping
from functools import partial
from cgi import parse_header


from promise import Promise
from sanic.response import HTTPResponse
from sanic.views import HTTPMethodView
from sanic_jwt.decorators import inject_user

from graphql.type.schema import GraphQLSchema
from graphql.execution.executors.asyncio import AsyncioExecutor
from graphql_server import (HttpQueryError, default_format_error,
                            encode_execution_results, json_encode,
                            load_json_body, run_http_query)

from loguru import logger
import backend.model as ax_model
import backend.auth as ax_auth
import backend.schema as ax_schema


start_time = None


def log_reguest(request, time=None):
    """ Logs graphql reuests """
    request_type = 'query'
    type_match = re.search('^(.*?) ', request.json['query'])
    if type_match:
        request_type = type_match.group(0)
    names_match = re.findall('([a-zA-Z].+?)\(', request.json['query'])

    if not names_match:
        # query without params
        without_match = re.findall('(.[a-zA-Z]+?) *?{', request.json['query'])
        command_names = without_match[0] + '...'
    else:
        clear_match = []
        for name in names_match:
            if 'query' not in name and 'mutation' not in name:
                clear_match.append(name)
        command_names = ', '.join(clear_match)

    # msg = request.json['query']
    msg = f'{request_type} -> {command_names}'
    # if time:
    #     msg += f' : {time}'

    logger.debug(msg)


class GraphQLView(HTTPMethodView):
    """ Local fork of sanic_graphql lib """

    decorators = [inject_user(), ax_auth.ax_protected()]

    schema = None
    executor = None
    root_value = None
    context = None
    pretty = False
    graphiql = False
    graphiql_version = None
    graphiql_template = None
    middleware = None
    batch = False
    jinja_env = None
    max_age = 86400

    _enable_async = True

    methods = ['GET', 'POST', 'PUT', 'DELETE']

    def __init__(self, **kwargs):
        """ Initiate """
        super(GraphQLView, self).__init__()
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

        self._enable_async = self._enable_async and isinstance(
            self.executor, AsyncioExecutor)
        assert isinstance(
            self.schema, GraphQLSchema), 'A Schema is required'

    # noinspection PyUnusedLocal

    def get_root_value(self, request):  # pylint: disable=unused-argument
        """ Some getter """
        return self.root_value

    def get_context(self, request):
        """ Returns saved context. Contains reqest, db_session, user """
        context = (
            self.context.copy()
            if self.context and
            isinstance(self.context, Mapping)
            else {}
        )
        if isinstance(context, Mapping) and 'request' not in context:
            context.update({'request': request})

        # log_reguest(request)
        # logger.debug(request.json['query'])

        return context

    def get_middleware(self, request):  # pylint: disable=unused-argument
        """ Some getter """
        return self.middleware

    def get_executor(self, request):  # pylint: disable=unused-argument
        """ Some getter """
        return self.executor

    format_error = staticmethod(default_format_error)
    encode = staticmethod(json_encode)

    async def dispatch_request(self, request, *args, **kwargs):
        """ Sanic view request dispatch.  """
        start_time = time.time()  # TODO this is debug profile

        with ax_model.scoped_session("GQL error -") as db_session:
            try:
                request_method = request.method.lower()
                data = self.parse_body(request)

                show_graphiql = request_method == 'get' and self.should_display_graphiql(
                    request)
                catch = show_graphiql

                pretty = self.pretty or show_graphiql or request.args.get(
                    'pretty')

                user = kwargs['user'] or None

                # auth_header = request.headers['authorization']
                # print(auth_header)

                ax_context = self.get_context(request)
                ax_context.update({'session': db_session})
                ax_context.update({'user': user})

                if request_method != 'options':
                    execution_results, all_params = run_http_query(
                        ax_schema.schema,
                        request_method,
                        data,
                        query_data=request.args,
                        batch_enabled=self.batch,
                        catch=catch,

                        # Execute options
                        return_promise=self._enable_async,
                        root_value=self.get_root_value(request),
                        context_value=ax_context,
                        middleware=self.get_middleware(request),
                        executor=self.get_executor(request),
                    )  # pylint: disable=unused-argument
                    del all_params
                    awaited_execution_results = await Promise.all(
                        execution_results)
                    result, status_code = encode_execution_results(
                        awaited_execution_results,
                        is_batch=isinstance(data, list),
                        format_error=self.format_error,
                        encode=partial(self.encode, pretty=pretty)
                    )
                    log_reguest(request, (time.time() - start_time))

                    return HTTPResponse(
                        result,
                        status=status_code,
                        content_type='application/json'
                    )
                else:
                    log_reguest(request, (time.time() - start_time))
                    return self.process_preflight(request)

            except HttpQueryError as err:
                logger.exception(f'graqlView -> {err}')
                return HTTPResponse(
                    self.encode({
                        'errors': [default_format_error(err)]
                    }),
                    status=err.status_code,
                    headers=err.headers,
                    content_type='application/json'
                )
            except Exception as err:    # pylint: disable=broad-except
                logger.exception(f'graqlView -> {err}')

    # noinspection PyBroadException

    def parse_body(self, request):
        """ - """
        content_type = self.get_mime_type(request)
        if content_type == 'application/graphql':
            return {'query': request.body.decode('utf8')}

        elif content_type == 'application/json':
            return load_json_body(request.body.decode('utf8'))

        elif content_type in (
                'application/x-www-form-urlencoded', 'multipart/form-data'):
            return request.form

        return {}

    @staticmethod
    def get_mime_type(request):
        """ - """
        # We use mimetype here since we don't need the other
        # information provided by content_type
        if 'content-type' not in request.headers:
            return None

        mimetype, _ = parse_header(request.headers['content-type'])
        return mimetype

    def should_display_graphiql(self, request):
        """ - """
        if not self.graphiql or 'raw' in request.args:
            return False

        return self.request_wants_html(request)

    def request_wants_html(self, request):
        """ - """
        accept = request.headers.get('accept', {})
        return 'text/html' in accept or '*/*' in accept

    def process_preflight(self, request):
        """ Preflight request support for apollo-client
        https://www.w3.org/TR/cors/#resource-preflight-requests """
        origin = request.headers.get('Origin', '')
        method = request.headers.get(
            'Access-Control-Request-Method', '').upper()

        if method and method in self.methods:
            return HTTPResponse(
                status=200,
                headers={
                    'Access-Control-Allow-Origin': origin,
                    'Access-Control-Allow-Methods': ', '.join(self.methods),
                    'Access-Control-Max-Age': str(self.max_age),
                }
            )
        else:
            return HTTPResponse(
                status=400,
            )



"""Ax GraphQL schema
1. Combines all mutations, queryes and subscribtions from schemas folder.
    Creates static part of schema from them.
2. Creates dynamic part of schema. For each AxGrid the dynamic schema is created
    Schema contains all fields of form. Rows are filtered using AxGris server
    filter. The name of schema is FormName+GridName. If grid is default, the
    additional schema is created, named FormName.
"""
import sys
import json
import graphene
from loguru import logger
from sqlalchemy import literal
from graphene_sqlalchemy import SQLAlchemyConnectionField
from dotmap import DotMap

from backend.schemas.users_schema import UsersQuery
from backend.schemas.users_schema import UsersMutations, UsersSubscription
from backend.schemas.home_schema import HomeQuery, HomeMutations
from backend.schemas.form_schema import FormQuery, FormMutations
from backend.schemas.workflow_schema import WorkflowMutations, WorkflowQuery
from backend.schemas.action_schema import ActionQuery, ActionMutations, \
    ActionSubscription
from backend.schemas.grids_schema import GridsQuery, GridsMutations
from backend.schemas.pages_schema import PagesQuery, PagesMutations
from backend.schemas.marketplace_schema import MarketplaceQuery,\
    MarketplaceMutations
from backend.schemas.fields_schema import FieldsQuery, FieldsMutations,\
    FieldsSubscription
from backend.schemas.types import Form, FieldType, Field, Grid, \
    Column, User, Role, State, Action, RoleFieldPermission, Group2Users, \
    Action2Role, State2Role, Role2Users, PositionInput, Page, Page2Users

from backend.model import AxForm
import backend.model as ax_model
import backend.dialects as ax_dialects
import backend.auth as ax_auth
import backend.exec as ax_exec


this = sys.modules[__name__]
schema = None

#  Dict used to determin what graphene type to use for AxFieldType.value_type
type_dictionary = {
    'text': graphene.String,
    'TEXT': graphene.String,
    'VARCHAR(255)': graphene.String,
    'INT': graphene.Int,
    'DECIMAL(65,2)': graphene.Float,
    'BOOL': graphene.Boolean,
    'GUID': graphene.String,
    'JSON': graphene.String,
    'TIMESTAMP': graphene.Int,
    'BLOB': graphene.String
}

# Static GQL types. Are same as SqlAlchemy model.
gql_types = [
    Form,
    FieldType,
    Field,
    Grid,
    Column,
    User,
    Role,
    State,
    Action,
    RoleFieldPermission,
    Group2Users,
    Action2Role,
    State2Role,
    Role2Users,
    Page,
    Page2Users,
    PositionInput
]

# Dynamic GQL types.
type_classes = {}


class Query(HomeQuery, FormQuery, UsersQuery, GridsQuery, WorkflowQuery,
            ActionQuery, PagesQuery, MarketplaceQuery, FieldsQuery,
            graphene.ObjectType):
    """Combines all schemas queryes"""
    # TODO - check if this fields are needed? Maybe custom queryes are enough?
    forms = SQLAlchemyConnectionField(Form)
    field_types = SQLAlchemyConnectionField(FieldType)
    fields = SQLAlchemyConnectionField(Field)
    grids = SQLAlchemyConnectionField(Grid)
    columns = SQLAlchemyConnectionField(Column)
    users = SQLAlchemyConnectionField(User)
    roles = SQLAlchemyConnectionField(Role)
    states = SQLAlchemyConnectionField(State)
    actions = SQLAlchemyConnectionField(Action)
    role_field_permissions = SQLAlchemyConnectionField(RoleFieldPermission)
    group_to_users = SQLAlchemyConnectionField(Group2Users)


class Mutations(HomeMutations, FormMutations, UsersMutations, GridsMutations,
                WorkflowMutations, PagesMutations, ActionMutations,
                MarketplaceMutations, FieldsMutations, graphene.ObjectType):
    """Combines all schemas mutations"""


class Subscription(UsersSubscription, ActionSubscription, FieldsSubscription,
                   graphene.ObjectType):
    """Combines all schemas subscription"""


async def exec_grid_code(form, grid, arguments=None, current_user=None):
    """ Execute python code to get SQL query for specific grid """

    db_fields_sql = ", ".join(
        '"' + field.db_name + '"' for field in form.db_fields)

    if db_fields_sql:
        db_fields_sql = ", " + db_fields_sql

    fields_sql = (f'"guid", "axState" {db_fields_sql}')

    localz = dict()
    ax = DotMap()  # javascript style dicts item['guid'] == item.guid
    ax.db_fields = fields_sql
    ax.db_table = form.db_name
    ax.row.guid = form.row_guid
    ax.arguments = arguments
    ax.user_email = None
    ax.user_guid = None
    if current_user:
        ax.user_email = current_user.get("email", None)
        ax.user_guid = current_user.get("user_id", None)
    ax.form = form
    ax.grid = grid
    ax.sql = ax_dialects.dialect.custom_query

    localz['ax'] = ax

    try:
        # exec(str(grid.code), globals(), localz)   # pylint: disable=exec-used
        await ax_exec.aexec(code=str(grid.code), localz=localz, ax=ax)
        ret_ax = localz['ax']
        ret_query = None
        if ret_ax and ret_ax.query:
            ret_query = str(ret_ax.query)
        return ret_query
    except SyntaxError as err:
        logger.exception(
            f'Error in query for [grid.name] - SyntaxError - {err}')
        return None
    except Exception as err:    # pylint: disable=broad-except
        logger.exception(
            f'Error in query for [grid.name] - SyntaxError - {err}')
        return None


def make_query_resolver(db_name, type_class):
    """ Dynamicly create resolver for GrapQL query based on
        db_name - db_name of AxForm + db_name of AxGrid.
        or only db_name of AxForm for default view grid

        Example:
        db_name=MyUberFormGridOne
        MyUberForm is form name
        GridOne is grid name

    """

    async def resolver(
            self, info, arguments=None, update_time=None, quicksearch=None,
            guids=None):
        del update_time, self
        err = f"Schema -> Resolver for {db_name}"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            current_user = info.context['user']
            # Find AxForm with name that db_name is started with
            ax_form = None
            found_forms = db_session.query(AxForm).filter(
                literal(db_name).startswith(AxForm.db_name)
            ).all()

            ax_grid = {}
            default_grid = {}

            for form in found_forms:
                # iterate all grids to know, what grid to use
                for grid in form.grids:
                    if form.db_name + grid.db_name == db_name:
                        ax_grid = grid
                        ax_form = form
                    if grid.is_default_view and form.db_name == db_name:
                        default_grid = grid
                        ax_form = form

            grid_to_use = ax_grid or default_grid
            if not grid_to_use:
                return None

            # TODO Add paging

            results = None

            if quicksearch or guids:
                # Quicksearch or 1tom query
                # TODO check if permissions needed !IMPORTANT
                results = await ax_dialects.dialect.select_all(
                    db_session=db_session,
                    ax_form=ax_form,
                    quicksearch=quicksearch,
                    guids=guids)
            else:
                # Normal gql query
                arguments_dict = None
                if arguments:
                    try:
                        arguments_dict = json.loads(arguments)
                    except Exception as err:
                        logger.exception(
                            f'Error reading arguments for grid query - {err}')
                        raise

                sql = None
                if grid_to_use.code:
                    sql = await exec_grid_code(
                        form=ax_form,
                        grid=grid_to_use,
                        arguments=arguments_dict,
                        current_user=current_user)
                else:
                    sql = ax_dialects.default_grid_query

                results = await ax_dialects.dialect.select_custom_query(
                    db_session=db_session,
                    ax_form=ax_form,
                    sql=sql,
                    arguments=arguments_dict)

            if not results:
                return None

            current_user = info.context['user']
            user_guid = None
            if current_user:
                user_guid = current_user["user_id"]

            allowed_fields = await ax_auth.get_grid_allowed_fields_dict(
                ax_form=ax_form,
                user_guid=user_guid)

            result_items = []
            for row in results:
                kwargs = {}
                row_state_name = row["axState"]
                for key, value in row.items():
                    if current_user and current_user['is_admin']:
                        kwargs[key] = await ax_dialects.dialect.conver_for_graphql(value)
                    elif (key in allowed_fields[row_state_name] and
                          allowed_fields[row_state_name][key] and
                          allowed_fields[row_state_name][key] > 0):
                        kwargs[key] = await ax_dialects.dialect.conver_for_graphql(value)
                    elif key == 'guid':
                        kwargs[key] = value
                    else:
                        kwargs[key] = None

                result_items.append(type_class(**kwargs))

            return result_items

    resolver.__name__ = f'resolve_{db_name}'
    return resolver


def make_to1_resolver(ax_field_guid, to1_field_db_name, related_form_db_name,
                      class_name):
    """ Creates resolver for to1 field. It gets value of field (guid)
        then makes sql query and returns right graphene field as query result"""

    async def resolver(parent, info):
        err = f"Error in {class_name} -> {ax_field_guid} resolver"
        with ax_model.try_catch(info.context['session'], err) as db_session:

            kwargs = {}
            field_value = getattr(parent, to1_field_db_name)
            if field_value:
                ax_form = db_session.query(AxForm).filter(
                    AxForm.db_name == related_form_db_name
                ).first()

                results = await ax_dialects.dialect.select_one(
                    db_session=db_session,
                    form=ax_form,
                    fields_list=ax_form.db_fields,
                    row_guid=field_value)

                if results:
                    for key, value in results[0].items():
                        kwargs[key] = await ax_dialects.dialect.conver_for_graphql(value)

                if class_name in type_classes:
                    return type_classes[class_name](**kwargs)
            return None

    resolver.__name__ = f'resolve_{to1_field_db_name}'
    return resolver


def make_tom_resolver(ax_field_guid, to1_field_db_name, related_form_db_name,
                      class_name):
    """ Creates resolver for tom field. It gets value of field
        (json with list of guids). then makes sql query and returns graphene
        List as query result"""

    async def resolver(parent, info):
        err = f"Error in {class_name} -> {ax_field_guid} resolver"
        with ax_model.try_catch(info.context['session'], err) as db_session:

            ret_result = []

            field_value = getattr(parent, to1_field_db_name)
            if field_value:
                guids_list = json.loads(field_value)
                ax_form = db_session.query(AxForm).filter(
                    AxForm.db_name == related_form_db_name
                ).first()

                items = {
                    "items": guids_list
                }

                results = await ax_dialects.dialect.select_all(
                    db_session=db_session,
                    ax_form=ax_form,
                    quicksearch=None,
                    guids=json.dumps(items))

                if results:
                    for row in results:
                        kwargs = {}
                        for key, value in row.items():
                            kwargs[key] = await ax_dialects.dialect.conver_for_graphql(value)
                        if class_name in type_classes:
                            row_class = type_classes[class_name](**kwargs)
                            ret_result.append(row_class)

            return ret_result

    resolver.__name__ = f'resolve_{to1_field_db_name}'
    return resolver


def get_class_name(form_db_name, grid_db_name):
    """ makes form db_name uppercase + grid db_name """
    capital_form_db_name = form_db_name[0].upper() + form_db_name[1:]
    class_name = capital_form_db_name + grid_db_name
    return class_name


def create_class_fields(form):
    """ Creates graphene fields for each AxField of form.
        If field is to1 or tom - creates resolver for that field
     """
    class_fields = {}
    class_fields['guid'] = graphene.String()
    class_fields['axState'] = graphene.String()
    class_fields['axLabel'] = graphene.String()

    # Add fields for each field of AxForm
    for field in form.db_fields:
        if (field.is_relation_field and field.options and
                field.options['form'] and field.options['grid']):

            form_db_name = field.options['form']
            grid_db_name = field.options['grid']
            class_name = get_class_name(form_db_name, grid_db_name)

            # if class_name in type_classes:
            
            # This is same as -> field_class = lambda: type_classes[class_name]
            # def field_class():
            #     return type_classes[class_name]
            field_class = lambda: type_classes[class_name]

            if field.is_to1_field:
                resolver = make_to1_resolver(
                    ax_field_guid=field.guid,
                    class_name=class_name,
                    related_form_db_name=form_db_name,
                    to1_field_db_name=field.db_name)

                class_fields[field.db_name] = graphene.Field(field_class)
                class_fields[f'resolve_{field.db_name}'] = resolver
            elif field.is_tom_field:
                resolver = make_tom_resolver(
                    ax_field_guid=field.guid,
                    class_name=class_name,
                    related_form_db_name=form_db_name,
                    to1_field_db_name=field.db_name)

                class_fields[field.db_name] = graphene.List(field_class)
                class_fields[f'resolve_{field.db_name}'] = resolver
        else:
            field_type = (
                type_dictionary[field.field_type.value_type])
            class_fields[field.db_name] = field_type()

    return class_fields





def init_schema(db_session):
    """Initiate GQL schema. Create dynamic part of schema and combine it with
    static part take from schemas folder"""
    error_msg = "Schema -> init_schema. Error initiating GraphQL shcema."
    with ax_model.try_catch(db_session, error_msg) as db_session:
        # Create typeClass based on each AxForm
        this.schema = None
        all_types = gql_types.copy()  # Add dynamic types to GQL schema

        ax_forms = db_session.query(AxForm).all()
        for form in ax_forms:
            for grid in form.grids:
                # For each grid we create Graphene field with name FormGrid
                capital_form_db_name = form.db_name[0].upper(
                ) + form.db_name[1:]
                class_name = capital_form_db_name + grid.db_name

                class_fields = create_class_fields(form)

                if class_fields:
                    # Create graphene class and append class dict
                    graph_class = type(
                        class_name,
                        (graphene.ObjectType,),
                        class_fields,
                        name=class_name,
                        description=form.name
                    )
                    type_classes[class_name] = graph_class

                    # if grid is default view we add enother class with name
                    # Form without Grid name
                    if grid.is_default_view is True:
                        default_class_name = form.db_name[0].upper(
                        ) + form.db_name[1:]

                        default_graph_class = type(
                            default_class_name,
                            (graphene.ObjectType,),
                            class_fields,
                            name=default_class_name,
                            description=form.name
                        )
                        type_classes[default_class_name] = default_graph_class


        # Dynamicly crate resolvers for each typeClass
        # Iterate throw created classes and create resolver for each
        dynamic_fields = {}
        for key, type_class in type_classes.items():
            all_types.append(type_class)
            dynamic_fields[key] = graphene.List(
                type_class,
                update_time=graphene.Argument(
                    type=graphene.String, required=False),
                quicksearch=graphene.Argument(
                    type=graphene.String, required=False),
                guids=graphene.Argument(
                    type=graphene.String, required=False),
                arguments=graphene.Argument(
                    type=graphene.String, required=False)
            )

            dynamic_fields['resolve_%s' % key] = make_query_resolver(
                key, type_class)

        # Combine static schema query and dynamic query
        DynamicQuery = type('DynamicQuery', (Query,), dynamic_fields)

        this.schema = graphene.Schema(
            query=DynamicQuery,
            mutation=Mutations,
            types=all_types,
            subscription=Subscription
        )


def init_schema_standalone():
    """ Initiate GQL schema without db_session """
    error_msg = "Schema -> init_schema. Error initiating GraphQL shcema."
    with ax_model.scoped_session(error_msg) as db_session:
        init_schema(db_session)



""" This is part of GraphQL schema (Mutaions, Queryes, Subscriptions).
Defines manipulation with Grids - create/update/delete of AxGrid and AxColumn
Mutation and queryes used only by constructor. AxGrid uses dynamic GQL schema.
Sea schema.py for more info.
"""
import uuid
import json
import graphene
from backend.model import AxForm, AxGrid, AxColumn, AxField
import backend.dialects as ax_dialects
import backend.model as ax_model


# import backend.cache as ax_cache # TODO use cache!
from backend.schemas.types import Grid, Column, PositionInput
from backend.auth import ax_admin_only


def get_default_grid_code():
    """ Returns default code for grid.
        Used in createTable (grids_schema)
        and createForm (home_schema)  """
    code = f"""ax.query = f\"\"\"
    {ax_dialects.default_grid_query}
\"\"\""""
    return code


def tom_sync_grid(db_session, form_db_name, old_db_name, new_db_name):
    """ Relation fields such as Ax1to1, Ax1tom and Ax1tomTable are using
    options_json to store db_name of grid used. If grid db_name is changed,
    the Json's in every relation field must be updated.
    For each relation field - check if options contain old form and old grid.

    Args:
        form_db_name (str): Current form db_name
        old_db_name (str): Old db_name of grid
        new_db_name (str): New db_name of grid
    """
    relation_fields = db_session.query(AxField).filter(
        AxField.field_type_tag.in_(('Ax1to1', 'Ax1tom', 'Ax1tomTable'))
    ).all()
    for field in relation_fields:
        if 'form' in field.options.keys() and 'grid' in field.options.keys():
            if field.options['grid'] == old_db_name and \
                    field.options['form'] == form_db_name:

                new_options = field.options
                new_options['grid'] = new_db_name

                field.options_json = json.dumps(new_options)
                db_session.flush()


class CreateColumn(graphene.Mutation):
    """ Creates AxColumn

    Arguments:
        grid_guid (str): Guid of AxGrid
        field_guid (str): Guid of AxField
        column_type (str): Not used!. Was used for responsive grid. #TODO delete
        position (int): Position of column in grid
        positions (List(PositionInput)): Position of other columns in grid

    Returns:
        column (AxColumn): Created grid column
    """
    class Arguments:  # pylint: disable=missing-docstring
        grid_guid = graphene.String()
        field_guid = graphene.String()
        column_type = graphene.String(required=False, default_value=None)
        position = graphene.Int(required=False, default_value=None)
        positions = graphene.List(
            PositionInput, required=False, default_value=None)

    ok = graphene.Boolean()
    column = graphene.Field(Column)

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'grids_schema -> CreateColumn'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            grid_guid = args.get('grid_guid')
            field_guid = args.get('field_guid')
            column_type = args.get('column_type')
            position = args.get('position')
            positions = args.get('positions')

            ax_grid = db_session.query(AxGrid).filter(
                AxGrid.guid == uuid.UUID(grid_guid)
            ).first()

            if not positions:
                positions = []
            if not position:
                position = len(ax_grid.columns)

            ax_column = AxColumn()
            ax_column.grid_guid = ax_grid.guid
            ax_column.field_guid = uuid.UUID(field_guid)
            ax_column.column_type = column_type
            ax_column.position = position
            db_session.add(ax_column)

            for column in ax_grid.columns:
                for pos in positions:
                    if str(column.guid) == pos.guid:
                        column.position = pos.position
                        column.column_type = pos.parent

            db_session.flush()
            return CreateColumn(column=ax_column, ok=True)


class UpdateColumnOptions(graphene.Mutation):
    """Updates options of column"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        options = graphene.String()

    ok = graphene.Boolean()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - grids_schema -> UpdateColumnOptions.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            ax_column = db_session.query(AxColumn).filter(
                AxColumn.guid == uuid.UUID(args.get('guid'))
            ).first()
            ax_column.options_json = args.get('options')
            db_session.flush()
            return UpdateColumnOptions(ok=True)


class DeleteColumn(graphene.Mutation):
    """ Deletes AxColumn """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - grids_schema -> DeleteColumn'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            ax_column = db_session.query(AxColumn).filter(
                AxColumn.guid == uuid.UUID(guid)
            ).first()
            db_session.delete(ax_column)
            return DeleteColumn(deleted=guid, ok=True)


class ChangeColumnsPositions(graphene.Mutation):
    """Change position and type of columns"""
    class Arguments:  # pylint: disable=missing-docstring
        grid_guid = graphene.String()
        positions = graphene.List(PositionInput)

    ok = graphene.Boolean()
    columns = graphene.List(Column)

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - grids_schema -> ChangeColumnsPositions.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            grid_guid = args.get('grid_guid')
            positions = args.get('positions')

            ax_grid = db_session.query(AxGrid).filter(
                AxGrid.guid == uuid.UUID(grid_guid)
            ).one()

            for column in ax_grid.columns:
                for pos in positions:
                    if column.guid == uuid.UUID(pos.guid):
                        column.position = pos.position
                        column.column_type = pos.parent
            db_session.flush()
            query = Column.get_query(info)
            column_list = query.filter(
                AxColumn.grid_guid == uuid.UUID(grid_guid)
            ).all()

            ok = True
            return ChangeColumnsPositions(columns=column_list, ok=ok)


class CreateGrid(graphene.Mutation):
    """ Creates AxGrid """
    class Arguments:  # pylint: disable=missing-docstring
        form_guid = graphene.String()
        name = graphene.String()

    ok = graphene.Boolean()
    grid = graphene.Field(Grid)

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        import backend.schema as ax_schema
        err = 'Error in gql mutation - grids_schema -> CreateGrid.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            form_guid = args.get('form_guid')
            name = args.get('name')

            ax_form = db_session.query(AxForm).filter(
                AxForm.guid == uuid.UUID(form_guid)
            ).first()

            db_name = "Grid"
            cur_num = 1
            name_is_checked = False
            while name_is_checked is False:
                error_flag = False
                if cur_num > 1:
                    cur_name = name + " " + str(cur_num)
                else:
                    cur_name = name
                cur_db_name = db_name + str(cur_num)
                for grid in ax_form.grids:
                    if grid.name == cur_name or grid.db_name == cur_db_name:
                        error_flag = True
                if error_flag is True:
                    cur_num = cur_num + 1
                else:
                    name_is_checked = True
                    break

            ax_grid = AxGrid()
            ax_grid.name = cur_name
            ax_grid.db_name = db_name + str(cur_num)
            ax_grid.form_guid = ax_form.guid
            ax_grid.code = None
            ax_grid.position = len(ax_form.grids) + 1

            default_options = {
                "enableQuickSearch": False,
                "enableFlatMode": False,
                "enableColumnsResize": True,
                "enableFiltering": True,
                "enableSorting": True,
                "enableOpenForm": True,
                "enableActions": True,
                "rowHeight": 45,
                "pinned": 0
            }

            ax_grid.options_json = json.dumps(default_options)
            ax_grid.is_default_view = False
            db_session.add(ax_grid)
            db_session.commit()
            ax_schema.init_schema(db_session)
            return CreateGrid(grid=ax_grid, ok=True)


class DeleteGrid(graphene.Mutation):
    """ Deletes AxFGrid """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        import backend.schema as ax_schema
        err = 'Error in gql mutation - grids_schema -> DeleteGrid.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')

            ax_grid = db_session.query(AxGrid).filter(
                AxGrid.guid == uuid.UUID(guid)
            ).first()
            form_db_name = ax_grid.form.name
            old_grid_db_name = ax_grid.db_name
            default_db_name = None

            for grid in ax_grid.form.grids:
                if grid.is_default_view:
                    default_db_name = grid.db_name

            db_session.delete(ax_grid)
            db_session.commit()
            ax_schema.init_schema(db_session)
            tom_sync_grid(db_session=db_session,
                          form_db_name=form_db_name,
                          old_db_name=old_grid_db_name,
                          new_db_name=default_db_name)
            return DeleteGrid(deleted=guid, ok=True)


class UpdateGrid(graphene.Mutation):
    """ Update AxGrid """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        name = graphene.String()
        db_name = graphene.String()
        options_json = graphene.String()
        is_default_view = graphene.Boolean()

    ok = graphene.Boolean()
    grid = graphene.Field(Grid)

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        import backend.schema as ax_schema
        err = 'Error in gql mutation - grids_schema -> UpdateGrid.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            name = args.get('name')
            db_name = args.get('db_name')
            options_json = args.get('options_json')
            is_default_view = args.get('is_default_view')
            schema_reload_needed = False
            tom_sync_needed = False

            ax_grid = db_session.query(AxGrid).filter(
                AxGrid.guid == uuid.UUID(guid)
            ).first()
            old_db_name = ax_grid.db_name

            if name:
                ax_grid.name = name

            if db_name:
                ax_grid.db_name = db_name
                schema_reload_needed = True
                if old_db_name != db_name:
                    tom_sync_needed = True

            if options_json:
                ax_grid.options_json = options_json

            if is_default_view:
                all_grids = db_session.query(AxGrid).filter(
                    AxGrid.form_guid == ax_grid.form_guid
                    and AxGrid.guid != ax_grid.guid
                ).all()

                for grid in all_grids:
                    grid.is_default_view = False

                ax_grid.is_default_view = is_default_view
                schema_reload_needed = True

            db_session.commit()

            if schema_reload_needed:
                ax_schema.init_schema(db_session)

            if tom_sync_needed:
                tom_sync_grid(
                    db_session=db_session,
                    form_db_name=ax_grid.form.db_name,
                    old_db_name=old_db_name,
                    new_db_name=ax_grid.db_name)

            return UpdateGrid(grid=ax_grid, ok=True)


class UpdateGridCode(graphene.Mutation):
    """ Update code of AxGrid """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        code = graphene.String()

    ok = graphene.Boolean()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - grids_schema -> UpdateGridCode.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            code = args.get('code')

            ax_grid = db_session.query(AxGrid).filter(
                AxGrid.guid == uuid.UUID(guid)
            ).first()

            if code is None or code == get_default_grid_code() or code == '':
                ax_grid.code = None
            else:
                ax_grid.code = code

            db_session.commit()
            return UpdateGridCode(ok=True)


class GridsQuery(graphene.ObjectType):
    """Query all data of AxGrid and related classes"""
    ax_grid = graphene.Field(
        Grid,
        form_db_name=graphene.Argument(type=graphene.String, required=True),
        grid_db_name=graphene.Argument(type=graphene.String, required=True),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )
    grids_list = graphene.List(
        Grid,
        form_db_name=graphene.Argument(type=graphene.String, required=True),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )

    async def resolve_ax_grid(
            self, info, form_db_name, grid_db_name, update_time=None):
        """Get AxGrid by form db_name and grid db_name"""
        del update_time
        err = 'grids_schema -> resolve_grid'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True) as db_session:
            grid = None
            ax_form = db_session.query(AxForm).filter(
                AxForm.db_name == form_db_name
            ).first()

            if ax_form:
                query = Grid.get_query(info=info)
                if grid_db_name and grid_db_name != 'null':
                    grid = query.filter(
                        AxGrid.form_guid == ax_form.guid
                    ).filter(
                        AxGrid.db_name == grid_db_name
                    ).first()
                else:
                    grid = query.filter(
                        AxGrid.form_guid == ax_form.guid
                    ).filter(
                        AxGrid.is_default_view.is_(True)
                    ).first()

                if not grid or not grid.code:
                    grid.code_not_none = get_default_grid_code()
                else:
                    grid.code_not_none = grid.code
            return grid

    @ax_admin_only
    async def resolve_grids_list(self, info, form_db_name):
        """Gets list of all AxGrid's of form """
        err = 'grids_schema -> resolve_grids_list'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True) as db_session:
            ax_form = db_session.query(AxForm).filter(
                AxForm.db_name == form_db_name
            ).first()

            query = Grid.get_query(info=info)
            grids = query.filter(AxGrid.form_guid == ax_form.guid).all()
            return grids


class GridsMutations(graphene.ObjectType):
    """Combine all mutations"""
    create_column = CreateColumn.Field()
    delete_column = DeleteColumn.Field()
    change_columns_positions = ChangeColumnsPositions.Field()
    create_grid = CreateGrid.Field()
    delete_grid = DeleteGrid.Field()
    update_grid = UpdateGrid.Field()
    update_grid_code = UpdateGridCode.Field()



"""Ax workflow apps builder.

Usage:
  ax [--host=<host>] [--port=<port>]
  ax --version
  ax --help

Options:
  -h --help       Show this screen.
  --version       Show version.
  --host=<host>   Host addres to run ax web application (default: 127.0.0.1).
  --port=<port>   Port of ax web application (default: 8080).

"""

import os
import sys
from pathlib import Path
from sanic import Sanic, response
from sanic_cors import CORS
from sanic_compress import Compress
from loguru import logger
from docopt import docopt

# fields specific
import stripe

# Add folder with main.py to sys.path.
root_path = Path(__file__).parent.resolve()
if root_path not in sys.path:
    sys.path.insert(0, str(root_path))


# pylint: disable=wrong-import-position
import backend.logger as ax_logger
import backend.misc as ax_misc
import backend.cache as ax_cache
import backend.model as ax_model
import backend.schema as ax_schema
import backend.pubsub as ax_pubsub
import backend.scheduler as ax_scheduler
import backend.migration as ax_migration
import backend.routes as ax_routes
import backend.dialects as ax_dialects
import backend.emails as ax_emails
import backend.auth as ax_auth

# pylint: enable=wrong-import-position


# def init_model():
#     """Initiate model. Used in alembic scripts"""
#     ax_model.init_model(
#         dialect=str(os.environ.get('AX_DB_DIALECT') or 'sqlite'),
#         host=str(os.environ.get('AX_DB_HOST') or '127.0.0.1'),
#         port=str(os.environ.get('AX_DB_PORT') or '5432'),
#         login=str(os.environ.get('AX_DB_LOGIN') or ''),
#         password=str(os.environ.get('AX_DB_PASSWORD') or ''),
#         database=str(os.environ.get('AX_DB_DATABASE') or 'ax'),
#         sqlite_filename=str(os.environ.get(
#             'AX_DB_SQLITE_FILENAME') or 'ax_sqlite.db'),
#         sqlite_absolute_path=os.environ.get(
#             'AX_DB_SQLITE_ABSOLUTE_PATH') or None
#     )


def init_fields():
    """ Initiate stripe. Used in AxPaymentStripe """
    if os.getenv('STRIPE_SECRET_KEY') and os.getenv('STRIPE_API_VERSION'):
        stripe.api_key = os.getenv('STRIPE_SECRET_KEY')
        stripe.api_version = os.getenv('STRIPE_API_VERSION')


def init_ax():
    """Initiate all modules of Ax"""

    # Load settings from app.yaml to os.environ
    ax_misc.load_configuration()
    # Misc functions used throw all application. Nothing specific.
    ax_misc.init_misc(
        timezone_name=str(os.environ.get('AX_TIMEZONE') or 'UTC'),
        tmp_absolute_path=os.environ.get('AX_TMP_ABSOLUTE_PATH') or None,
        uploads_absolute_path=os.environ.get(
            'AX_UPLOADS_ABSOLUTE_PATH') or None
    )
    # Logger (loguru) used in all modules - console + file + sentry
    ax_logger.init_logger(
        logs_filename=os.environ.get('AX_LOGS_FILENAME') or None,
        logs_absolute_path=os.environ.get('AX_LOGS_ABSOLUTE_PATH') or None,
        logs_level=os.environ.get('AX_LOGS_LEVEL') or 'INFO'
    )

    # for item, value in os.environ.items():
    #     logger.info('ENV: {}: {}'.format(item, value))

    # Initiate SqlAlchemy. Made with separate function for alembic.
    # It initiates database connection on migration.
    # init_model()
    ax_model.init_default_model()

    # Init cache module (aiocache). It is not yet used.
    ax_cache.init_cache(
        mode=str(os.environ.get('AX_CACHE_MODE') or 'default'),
        redis_endpoint=str(os.environ.get('AX_REDIS_ENDPOINT') or '127.0.0.1'),
        redis_port=int(os.environ.get('AX_REDIS_PORT') or 6379),
        redis_timeout=int(os.environ.get('AX_REDIS_TIMEOUT') or 1),
    )
    # Initiate pub-sub module. Used for web-socket subscriptions.
    # Notification on workflow action
    ax_pubsub.init_pubsub()
    # Check if database schema needs update
    ax_migration.init_migration()
    # Initiate gql schema.  Depends on cache and pubsub
    ax_schema.init_schema_standalone()

    # cors = CORS(app, resources={r"/api/*": {"origins": "*"}})
    CORS(app, automatic_options=True)  # TODO limit CORS to api folder

    # Enable gzip compression for responses
    # 'text/html','text/css','text/xml','application/json','application/java
    Compress(app)

    @app.listener('before_server_start')
    async def initialize_scheduler(_app, _loop):  # pylint: disable=unused-variable
        """Initiate scheduler. It used for cron-like jobs"""
        ax_scheduler.init_scheduler()

    @app.listener('before_server_start')
    def save_loop(_app, _loop):  # pylint: disable=unused-variable
        """Initiate graphene graphql server"""
        ax_routes.loop = _loop
        ax_routes.app = _app
        ax_routes.init_graphql_view()
        # _app.add_route(ax_routes.graphql_view, '/api/graphql')

    ssl_cert = os.environ.get('SSL_CERT_ABSOLUTE_PATH', None)
    ssl_key = os.environ.get('SSL_KEY_ABSOLUTE_PATH', None)
    ssl_enabled = False
    if ssl_cert and ssl_key:
        ssl_enabled = True

    if ssl_enabled:
        @app.middleware('request')
        async def force_ssl(request):     # pylint: disable=unused-variable
            if request.scheme == 'ws':
                return response.redirect(
                    request.url.replace('wss://', 'https://', 1),
                    status=301
                )
            elif request.scheme == 'http':
                return response.redirect(
                    request.url.replace('http://', 'https://', 1),
                    status=301
                )

    # Initiate all sanic server routes
    ax_routes.init_routes(
        sanic_app=app,
        pages_path=os.environ.get('AX_PAGES_ABSOLUTE_PATH'),
        ssl_enabled=ssl_enabled
    )
    # Initiate SQL dialects module. Different SQL queries for differents DBs
    ax_dialects.init_dialects(os.environ.get('AX_DB_DIALECT') or 'sqlite')

    # Initiate email sender
    ax_emails.init_email(
        smtp_host=os.environ.get('AX_SMTP_HOST'),
        smtp_login=os.environ.get('AX_SMTP_LOGIN'),
        smtp_password=os.environ.get('AX_SMTP_PASSWORD'),
        port=os.environ.get('AX_SMTP_PORT'),
        ssl=os.environ.get('AX_SMTP_SSL')
    )

    # Initiate auth module
    ax_auth.init_auth(
        sanic_app=app, secret=os.environ.get('AX_AUTH_SECRET'))

    init_fields()


ax_logo = """

                                           .
                                          .(
       ..                                 /(#
       ...           --- AX ---          ,/((////.
       ...,,                             ***(/////,
     ****,,,*/                          (****((///*
      ****///////                   ./(((#***/((/*,,,.
      ,/**////////((((/,      ,(((((((((/((**,.,..
        # (///((((((((((((#((((((((((////* ..
          # (((((((((((((##((((((((((((///,
             .#(((((((  ##(((((((((((((///
                       # (###/      /(///
                      # (//,
                      # (#            (//

"""


def main():
    """Main function"""
    arguments = docopt(__doc__)

    host = os.environ.get('AX_HOST', '127.0.0.1')
    port = int(os.environ.get('AX_PORT', 80))
    debug = bool(os.environ.get('AX_SANIC_DEBUG', False))
    access_log = bool(os.environ.get('AX_SANIC_ACCESS_LOG', False))
    workers = int(os.environ.get('AX_SANIC_WORKERS', 1))
    ssl_cert = os.environ.get('SSL_CERT_ABSOLUTE_PATH', None)
    ssl_key = os.environ.get('SSL_KEY_ABSOLUTE_PATH', None)

    if arguments['--host']:
        host = arguments['--host']

    if arguments['--port']:
        port = int(arguments['--port'])

    ssl = None
    protocol = 'http'
    if ssl_cert and ssl_key:
        ssl = {'cert': ssl_cert, 'key': ssl_key}
        protocol = 'https'

    print(ax_logo)
    logger.info(
        f'Ax is running with {workers} workers. Admin is avalible - {protocol}://{host}:{port}/admin/signin\n')
    app.run(
        host=host,
        port=port,
        debug=debug,
        access_log=access_log,
        workers=workers,
        ssl=ssl)


app = Sanic("Ax_workflow")
init_ax()

if __name__ == "__main__":
    main()



""" Auth module based on sanic-jwt """

import os
import sys
import uuid
import functools
import datetime
import asyncio
from loguru import logger
from passlib.hash import pbkdf2_sha256
from sanic_jwt import exceptions, initialize, Configuration
from sanic_jwt.decorators import _do_protection

from backend.model import AxUser, AxGroup2Users, AxRole, AxRole2Users, \
    AxRoleFieldPermission, AxForm, AxAction2Role, AxField
import backend.cache as ax_cache
import backend.model as ax_model
import backend.misc as ax_misc
import backend.dialects as ax_dialects
import backend.exec as ax_exec
import backend.fields.AxAuthor as AxFieldAxAuthor  # pylint: disable=unused-import

this = sys.modules[__name__]


async def get_state_guid(ax_form, state_name):
    """ Returns guid of state by name.
        If no state_name provided -> returns guid of Start state """
    state_guid = None
    for state in ax_form.states:
        if state_name and state.name == state_name:
            state_guid = str(state.guid)
        elif ((state_name is None or state_name == '') and
              state.is_start is True):
            state_guid = str(state.guid)

    return state_guid


async def check_if_action_avalible(
        action,
        current_user,
        dynamic_role_guids=None):
    """ Checks action workflow perms and dynamic roles perms"""
    user_is_admin = current_user.get(
        'is_admin', False) if current_user else False
    user_guid = current_user.get('user_id', None) if current_user else None

    action_guid = str(action.guid)
    avalible = await check_action_perm(
        user_guid=user_guid, action_guid=action_guid)

    if dynamic_role_guids:
        for role_guid in dynamic_role_guids:
            if not avalible:
                avalible = await check_action_perm(
                    user_guid=role_guid, action_guid=action_guid)

    if avalible or user_is_admin:
        return True
    return False


async def check_action_perm(user_guid, action_guid):
    """ Checks for action_perm_ cache record. """
    user_guid_str = str(user_guid)
    action_guid_str = str(action_guid)
    action_key = f"action_perm_{user_guid_str}_{action_guid_str}"
    perm = await ax_cache.cache.get(action_key)
    return perm


async def get_allowed_fields_dict(ax_form, user_guid, state_guid):
    """ Gets allowed fields for user for specific state
         -> dict['field_guid'] = 1/2  ; 1 - read, 2 - edit"""
    cache_keys = []
    field_guids = []
    allowed_fields_dict = {}
    for field in ax_form.fields:
        user_guid = str(user_guid)
        field_guid = str(field.guid)
        state_guid = str(state_guid)
        key = f"perm_{user_guid}_{field_guid}_{state_guid}"
        cache_keys.append(key)
        field_guids.append(str(field.guid))

    cache_results = await ax_cache.cache.multi_get(cache_keys)

    for idx, field_guid in enumerate(field_guids):
        perm_value = cache_results[idx]
        allowed_fields_dict[field_guid] = perm_value

    return allowed_fields_dict


async def check_field_perm(db_session, current_user, field_guid, row_guid):
    """ Checks if user can view or modify for specific field and row
        returns None = forbidden, 1 = read, 2 = edit"""

    user_guid = current_user.get("user_id", None) if current_user else None
    user_id_admin = current_user.get(
        "is_admin", False) if current_user else False

    if user_id_admin:
        return 2

    ax_field = db_session.query(AxField).filter(
        AxField.guid == uuid.UUID(str(field_guid))
    ).first()

    result = await ax_dialects.dialect.select_one(
        db_session=db_session,
        form=ax_field.form,
        fields_list=[],
        row_guid=row_guid)

    state_name = result[0]['axState']
    state_guid = await get_state_guid(
        ax_form=ax_field.form, state_name=state_name)
    key = f"perm_{user_guid}_{ax_field.guid}_{state_guid}"
    perm = await ax_cache.cache.get(key)
    return perm


async def get_grid_allowed_fields_dict(ax_form, user_guid):
    """ Gets allowed fields for user for all states
            -> dict['state_name']['field_db_name'] = 1/2 """
    cache_keys = []
    for state in ax_form.states:
        for field in ax_form.no_tab_fields:
            user_guid = str(user_guid)
            field_guid = str(field.guid)
            state_guid = str(state.guid)
            key = f"perm_{user_guid}_{field_guid}_{state_guid}"
            cache_keys.append(key)

    cache_results = await ax_cache.cache.multi_get(cache_keys)

    allowed_fields_dict = {}
    idx = 0
    for state in ax_form.states:
        allowed_fields_dict[state.name] = {}
        for field in ax_form.no_tab_fields:
            field_db_name = field.db_name
            if field.is_virtual:
                field_db_name = field.field_type.virtual_source
            allowed_fields_dict[state.name][field_db_name] = cache_results[idx]
            idx += 1
    return allowed_fields_dict


async def get_dynamic_roles_guids(ax_form, current_user):
    """ Get list of role guids of all dynamic roles of form.
        For each dynamic role - check if form values and user fits role """
    role_guids = []
    for role in ax_form.roles:
        if role.is_dynamic:
            # check if user and form fits dynamic role.
            form_fits_role = await check_fits_dynamic_role(
                role=role,
                ax_form=ax_form,
                current_user=current_user)
            if form_fits_role is True:
                role_guids.append(str(role.guid))
    return role_guids


async def check_fits_dynamic_role(role, ax_form, current_user):
    """ Runs check_dynamic_role() of field_type py file """
    if role.is_dynamic:
        fits = await ax_exec.execute_field_code(
            code=role.code,
            form=ax_form,
            current_user=current_user)
        if 'result' in fits and fits['result'] is True:
            return True
    return False


async def set_form_visibility(
        ax_form,
        state_guid,
        current_user,
        dynamic_role_guids=None):
    """ Sets permission field visibility. Deletes value if field is hidden """
    user_guid = current_user.get("user_id", None) if current_user else None
    user_id_admin = current_user.get(
        "is_admin", False) if current_user else False
    allowed_fields = await get_allowed_fields_dict(
        ax_form=ax_form,
        user_guid=user_guid,
        state_guid=state_guid)

    # Apply worklfow perms
    for field in ax_form.no_tab_fields:
        if field in ax_form.db_fields or field.field_type.is_virtual:
            field_guid_str = str(field.guid)
            field.is_hidden = True
            if user_id_admin:
                field.is_readonly = False
                field.is_hidden = False
            else:
                if field_guid_str in allowed_fields:
                    if allowed_fields[field_guid_str] == 1:
                        field.is_readonly = True
                        field.is_hidden = False
                    if allowed_fields[field_guid_str] == 2:
                        field.is_readonly = False
                        field.is_hidden = False

    # Apply dynamic role perms (rewrite previos perms)
    if dynamic_role_guids:
        for role_guid in dynamic_role_guids:
            allowed_role_fields = await get_allowed_fields_dict(
                ax_form=ax_form,
                user_guid=role_guid,
                state_guid=state_guid)

            for field in ax_form.no_tab_fields:
                if field in ax_form.db_fields or field.field_type.is_virtual:
                    field_guid_str = str(field.guid)
                    if field_guid_str in allowed_role_fields:
                        if allowed_role_fields[field_guid_str] == 1:
                            field.is_readonly = True
                            field.is_hidden = False
                        if allowed_role_fields[field_guid_str] == 2:
                            field.is_readonly = False
                            field.is_hidden = False
                            field.needs_sql_update = True

    for field in ax_form.db_fields:
        if not field.field_type.is_updated_always:
            # If field is hidden -> Epty the value and forbid
            # inser and update
            if field.is_hidden is True:
                field.value = None
                field.needs_sql_update = False

            # Forbid update if readonly
            if field.is_readonly is True:
                field.needs_sql_update = False

    return ax_form


async def check_if_admin(user_guid, db_session):
    """ Checks if user is adimn and writes  <user_is_admin_{user_guid}> cache"""
    err = "Error -> Auth -> check_if_admin"
    with ax_model.try_catch(db_session, err):
        # Get admin group, check if
        admin_group = db_session.query(AxUser).filter(
            AxUser.is_admin.is_(True)
        ).filter(
            AxUser.is_group.is_(True)
        ).first()

        group2user = db_session.query(AxGroup2Users).filter(
            AxGroup2Users.group_guid == admin_group.guid
        ).filter(
            AxGroup2Users.user_guid == uuid.UUID(user_guid)
        ).first()

        if group2user:
            key = f'user_is_admin_{user_guid}'
            await ax_cache.cache.set(key, True)


async def write_perm_cache(db_session, user_guid):
    """ On each user token refresh we write to cache permissions for
        each field/state combination """
    perm_cache_pairs = []
    # debug_list = []
    user_and_groups = []

    # Add 'everyone' group
    everyone_group = db_session.query(AxUser).filter(
        AxUser.is_everyone.is_(True)
    ).first()
    if everyone_group:
        user_and_groups.append(everyone_group.guid)

    if user_guid:
        # Add current user
        user_and_groups.append(uuid.UUID(user_guid))

        # Add all-users group
        # TODO get this from cache!
        all_users_group = db_session.query(AxUser).filter(
            AxUser.is_all_users.is_(True)
        ).first()
        user_and_groups.append(all_users_group.guid)

        # Get user groups
        group2user_result = db_session.query(AxGroup2Users).filter(
            AxGroup2Users.user_guid == uuid.UUID(user_guid)
        ).all()
        for g2u in group2user_result:
            if g2u.group_guid not in user_and_groups:
                user_and_groups.append(g2u.group_guid)

    # Get roles of user and his groups
    roles_guids = []
    role2user_result = db_session.query(AxRole2Users).filter(
        AxRole2Users.user_guid.in_(user_and_groups)
    ).all()
    for r2u in role2user_result:
        if r2u.role_guid not in roles_guids:
            roles_guids.append(r2u.role_guid)

    # Get AxRoleFieldPermission for those roles
    forms_guids = []
    perms = db_session.query(AxRoleFieldPermission).filter(
        AxRoleFieldPermission.role_guid.in_(roles_guids)
    ).all()
    for perm in perms:
        if perm.form_guid not in forms_guids:
            forms_guids.append(perm.form_guid)

    # Get AxAction2Role for those roles
    actions_guids = []
    action_role_result = db_session.query(AxAction2Role).filter(
        AxAction2Role.role_guid.in_(roles_guids)
    ).all()
    for a2r in action_role_result:
        if a2r.action_guid not in actions_guids:
            actions_guids.append(a2r.action_guid)
            action_guid = str(a2r.action_guid)
            action_key = f"action_perm_{user_guid}_{action_guid}"
            perm_cache_pairs.append([action_key, True])

    # Get all forms with perms
    all_forms = db_session.query(AxForm).filter(
        AxForm.guid.in_(forms_guids)
    ).all()

    # For each form
    for ax_form in all_forms:
        # For each field
        for ax_field in ax_form.no_tab_fields:
            # For each state
            for ax_state in ax_form.perm_states:
                read = False
                edit = False
                # For each perm
                form_perms = [p for p in perms if (
                    p.form_guid == ax_form.guid and
                    p.state_guid == ax_state.guid)]
                for perm in form_perms:
                    # if perm is set to whole form
                    # if perm is set to tab
                    # if perm is set to field

                    if (perm.field_guid is None or
                            perm.field_guid == ax_field.parent or
                            perm.field_guid == ax_field.guid):

                        if perm.read is True:
                            read = True
                        if perm.edit is True:
                            edit = True

                # Write cache perm_<user_guid>_<field_guid>_<state_guid> = 0/1/2
                field_guid = str(ax_field.guid)
                state_guid = str(ax_state.guid)
                key = f"perm_{user_guid}_{field_guid}_{state_guid}"
                value = 0
                if read is True:
                    value = 1
                if edit is True:
                    value = 2

                perm_cache_pairs.append([key, value])
                # debug_list.append(
                #     (f"{ax_form.db_name} -> "
                #      f"{ax_state.name} -> {ax_field.db_name} -> {value}")
                # )

    expire_seconds = 60 * 20  # 20 mins #TODO check this timeout
    if user_guid is None:
        expire_seconds = 0  # disable ttl for Everyone group
    await ax_cache.cache.multi_set(perm_cache_pairs, ttl=expire_seconds)


async def write_dynamic_roles_cache(db_session):
    """ Write cache for each field/state combination
        perm_{dynamic_role_guid}_{field_guid}_{state_guid}
        Write cache needed on start and after each perm edit, if role is dynamic
        """
    perm_cache_pairs = []
    debug_list = []

    # Get all dynamic roles
    roles_guids = []
    dynamic_roles = db_session.query(AxRole).filter(
        AxRole.is_dynamic.is_(True)
    ).all()
    for role in dynamic_roles:
        roles_guids.append(role.guid)

    # Get AxRoleFieldPermission for those roles
    forms_guids = []
    perms = db_session.query(AxRoleFieldPermission).filter(
        AxRoleFieldPermission.role_guid.in_(roles_guids)
    ).all()
    for perm in perms:
        if perm.form_guid not in forms_guids:
            forms_guids.append(perm.form_guid)

    # Get AxAction2Role for those roles
    action_role_result = db_session.query(AxAction2Role).filter(
        AxAction2Role.role_guid.in_(roles_guids)
    ).all()
    actions_guids = []
    for a2r in action_role_result:
        if a2r.action_guid not in actions_guids:
            actions_guids.append(a2r.action_guid)
            action_guid = str(a2r.action_guid)
            role_guid = str(a2r.role_guid)
            action_key = f"action_perm_{role_guid}_{action_guid}"
            perm_cache_pairs.append([action_key, True])

    # Get all forms with perms
    all_forms = db_session.query(AxForm).filter(
        AxForm.guid.in_(forms_guids)
    ).all()

    # For each form
    for ax_form in all_forms:
        # For each field
        for ax_field in ax_form.no_tab_fields:
            # For each state
            for ax_state in ax_form.perm_states:
                read = False
                edit = False
                # For each dynamic role
                for dyn_role in dynamic_roles:
                    role_perms = [p for p in perms if (
                        p.role_guid == dyn_role.guid and
                        p.state_guid == ax_state.guid)]

                    for perm in role_perms:
                        # if perm is set to whole form
                        # if perm is set to tab
                        # if perm is set to field
                        if (perm.field_guid is None or
                                perm.field_guid == ax_field.parent or
                                perm.field_guid == ax_field.guid):

                            if perm.read is True:
                                read = True
                            if perm.edit is True:
                                edit = True

                    # Cache perm_<role_guid>_<field_guid>_<state_guid> = 0/1/2
                    role_guid = str(dyn_role.guid)
                    field_guid = str(ax_field.guid)
                    state_guid = str(ax_state.guid)
                    key = f"perm_{role_guid}_{field_guid}_{state_guid}"
                    value = 0
                    if read is True:
                        value = 1
                    if edit is True:
                        value = 2

                    perm_cache_pairs.append([key, value])
                    debug_list.append(
                        (f"{dyn_role.name} -> {ax_form.db_name} -> "
                         f"{ax_state.name} -> {ax_field.db_name} -> {value}"))

    expire_seconds = 0  # does not depricate
    await ax_cache.cache.multi_set(perm_cache_pairs, ttl=expire_seconds)


async def write_info_cache(user):
    """ Save short_name and email in cache """
    name_key = f'user_short_name_{user.guid}'
    await ax_cache.cache.set(name_key, user.short_name)
    email_key = f'user_email_{user.guid}'
    await ax_cache.cache.set(email_key, user.email)


async def authenticate(request, *args, **kwargs):
    """ - """
    del args, kwargs
    msg = "Error -> Auth -> authenticate"
    with ax_model.scoped_session(msg) as db_session:
        email = request.json.get("email", None)
        password = request.json.get("password", None)

        if not email or not password:
            raise exceptions.AuthenticationFailed("Missing email or password.")

        user = db_session.query(AxUser).filter(
            AxUser.email == email
        ).filter(
            AxUser.is_blocked.is_(False)
        ).first()

        if user is None:
            raise exceptions.AuthenticationFailed("User not found.")

        if not pbkdf2_sha256.verify(password, user.password):
            raise exceptions.AuthenticationFailed("Password is incorrect.")

        await check_if_admin(user_guid=str(user.guid), db_session=db_session)
        await write_perm_cache(db_session=db_session, user_guid=str(user.guid))
        await write_info_cache(user)

        db_session.expunge(user)
        return user


def ax_protected(initialized_on=None, **kw):
    """ Wrapper for @protected sanic_jwt decorator.
    If not authorisation token is present - then it is anon user, do not send
    401 error  """
    def _aux(view):
        @functools.wraps(view)
        async def _inner(request, *args, **kwargs):
            # If there is no jwt headers -> then this is everyone user.
            # We must not protect route
            if 'authorization' not in request.headers:
                return await view(request, *args, **kwargs)
            if not "Bearer" in request.headers['authorization']:
                return await view(request, *args, **kwargs)

            kwargs.update({
                'initialized_on': initialized_on,
                'kw': kw,
                'request': request,
                'f': view,
            })
            return await _do_protection(*args, **kwargs)
        return _inner
    return _aux


def ax_admin_only(func):
    """ Decorator to check if user is admin """
    async def check_admin(*args, **kwargs):
        user = args[1].context["user"]
        if user and user["is_admin"] is True:
            return await func(*args, **kwargs)
        else:
            logger.error('Only for admins error')
            raise Exception('Only for admins')
    return check_admin


async def retrieve_user(request, payload, *args, **kwargs):
    """ Get user info. This info is transfered into routes with inject_user """
    del request, args, kwargs
    if payload:
        user_id = payload.get('user_id') or None

        if not ax_misc.string_is_guid(user_id):
            return None

        email = await ax_cache.cache.get(f'user_email_{user_id}')

        if not email:
            msg = "Auth -> retrieve_user"
            with ax_model.scoped_session(msg) as db_session:
                user = db_session.query(AxUser).filter(
                    AxUser.guid == ax_misc.guid_or_none(user_id)
                ).first()

                if user is not None:
                    # raise exceptions.AuthenticationFailed("User not found.")
                    await check_if_admin(
                        user_guid=user_id, db_session=db_session)
                    await write_perm_cache(
                        db_session=db_session, user_guid=user_id)
                    await write_info_cache(user)

        email = await ax_cache.cache.get(f'user_email_{user_id}')
        short_name = await ax_cache.cache.get(f'user_short_name_{user_id}')
        is_admin = await ax_cache.cache.get(f'user_is_admin_{user_id}')

        user = {
            "user_id": str(user_id),
            "is_admin": is_admin,
            "short_name": short_name,
            "email": email
        }
        return user
    else:
        return None


async def store_refresh_token(request, user_id, refresh_token, *args, **kwargs):
    """ Store refresh token in cache and write all perms cache for user """
    del args, kwargs
    device_guid = request.json['deviceGuid']
    key = f'refresh_token_{device_guid}_{user_id}'
    await ax_cache.cache.set(key, refresh_token)


async def retrieve_refresh_token(request, user_id, *args, **kwargs):
    """ Get refresh token from cache and refresh all perms cache for user """
    del args, kwargs
    device_guid = request.json['deviceGuid']
    key = f'refresh_token_{device_guid}_{user_id}'
    refresh_token = await ax_cache.cache.get(key)

    if refresh_token:
        with ax_model.scoped_session("Auth.store_refresh_token") as db_session:
            await check_if_admin(user_guid=user_id, db_session=db_session)
            await write_perm_cache(db_session=db_session, user_guid=user_id)

    return refresh_token


class AxConfiguration(Configuration):
    """ sanic-jwt config class """
    url_prefix = '/api/auth'


client_guid = None
user_num = None
expire_date = None


def apply_lise(db_session):
    """
        Copyright (C) 2020 Mikhail Marenov - All Rights Reserved
        You MAY NOT CHANGE source code of Ax workflow without writen permission
        author. If you change source code in order to activate PRO features -
        YOU MAY BE SUBJECT TO HEAVY CIVIL PENALTIES. THESE INCLUDE MONETARY
        DAMAGES, COURT COSTS, AND ATTORNEYS FEES INCURRED
        Please read LICENSE.md for more information.
    """
    import rsa
    import base64

    license_text = os.environ.get('AX_LICENSE', None)

    if license_text:
        public_key_str = (
            f'-----BEGIN RSA PUBLIC KEY-----\nMEgCQQCKDk8gSjXCh9R8VFKUMFhLxqx/'
            f'mkWDIUd9uMrHsbYs89t5GnMCoNkmL9nj\nazDQ2nyRxnVaYgJ2hFM49EomRFGVAg'
            f'MBAAE=\n-----END RSA PUBLIC KEY-----\n')
        public_key = rsa.PublicKey.load_pkcs1(public_key_str, 'PEM')
        data, signature = license_text.split('##')
        cl_guid, expire_str, users_number = data.split('#')
        expire = datetime.datetime.fromtimestamp(int(expire_str))
        sign = base64.b64decode(signature)

        time_delta = expire - datetime.datetime.utcnow()
        if time_delta.days <= 0:
            err = "License is expired!"
            logger.exception(err)
            raise Exception(err)

        try:
            rsa.verify(data.encode('utf-8'), sign, public_key)
        except rsa.VerificationError:
            logger.exception("License is not verified!")
            raise

        logger.info(f'License applied for {users_number} users.')
        this.client_guid = cl_guid
        this.user_num = int(users_number)
        this.expire = expire
    else:
        this.client_guid = None
        this.user_num = 5000
        this.expire = None

    number_of_users = db_session.query(AxUser).filter(
        AxUser.is_group.is_(False)
    ).filter(
        AxUser.is_blocked.is_(False)
    ).count()
    if number_of_users > this.user_num:
        err = f"Maximum number of users exceeded! Max={this.user_num}"
        logger.exception(err)
        raise Exception(err)


def lise_is_active():
    """
        Copyright (C) 2020 Mikhail Marenov - All Rights Reserved
        You MAY NOT CHANGE source code of Ax workflow without writen permission
        author. If you change source code in order to activate PRO features -
        YOU MAY BE SUBJECT TO HEAVY CIVIL PENALTIES. THESE INCLUDE MONETARY
        DAMAGES, COURT COSTS, AND ATTORNEYS FEES INCURRED
        Please read LICENSE.md for more information.
    """
    if this.client_guid:
        time_delta = this.expire - datetime.datetime.utcnow()
        if time_delta.days <= 0:
            err = "License is expired!"
            logger.exception(err)
            raise Exception(err)
        return True
    return False


def init_auth(sanic_app, secret="This is big secret, set me in app.yaml"):
    """
    Initiate sanic-jwt module

    Copyright (C) 2020 Mikhail Marenov - All Rights Reserved
    You MAY NOT CHANGE source code of Ax workflow without writen permission
    author. If you change source code in order to activate PRO features -
    YOU MAY BE SUBJECT TO HEAVY CIVIL PENALTIES. THESE INCLUDE MONETARY
    DAMAGES, COURT COSTS, AND ATTORNEYS FEES INCURRED
    Please read LICENSE.md for more information.
    """

    delta = 60  # seconds
    initialize(sanic_app,
               authenticate=authenticate,
               configuration_class=AxConfiguration,
               refresh_token_enabled=True,
               store_refresh_token=store_refresh_token,
               retrieve_refresh_token=retrieve_refresh_token,
               retrieve_user=retrieve_user,
               expiration_delta=delta,
               cookie_access_token_name='ax_auth',
               cookie_set=True,
               cookie_strict=False,
               login_redirect_url='/signin',
               secret=secret)

    with ax_model.scoped_session("init_auth - ERROR") as db_session:
        apply_lise(db_session)

        # Write cache form Everyone group
        asyncio.get_event_loop().run_until_complete(write_perm_cache(
            db_session=db_session, user_guid=None))

        # Write cache for dynamic roles
        asyncio.get_event_loop().run_until_complete(write_dynamic_roles_cache(
            db_session=db_session))



"""Defines Users Scheme and all mutations"""

import uuid
import json
import ast
import graphene
from sqlalchemy import or_
from graphene_sqlalchemy.converter import convert_sqlalchemy_type
from passlib.hash import pbkdf2_sha256
from loguru import logger

from backend.misc import convert_column_to_string
from backend.model import AxUser, GUID, AxGroup2Users, AxRole2Users
import backend.model as ax_model
import backend.cache as ax_cache
import backend.dialects as ax_dialects
# import backend.pubsub as ax_pubsub
import backend.auth as ax_auth
from backend.schemas.types import User
from backend.auth import ax_admin_only
import backend.misc as ax_misc

convert_sqlalchemy_type.register(GUID)(convert_column_to_string)


class CreateUser(graphene.Mutation):
    """ Creates AxUser """
    class Arguments:  # pylint: disable=missing-docstring
        email = graphene.String()
        name = graphene.String()
        short_name = graphene.String()
        password = graphene.String()
        avatar_tmp = graphene.String()
        info = graphene.String()
        is_blocked = graphene.Boolean()

    ok = graphene.Boolean()
    user = graphene.Field(User)
    msg = graphene.String()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        # avatar_tmp = args.get('avatar_tmp')

        err = 'Error in gql mutation - users_schema -> CreateUser.'
        with ax_model.try_catch(info.context['session'], err) as db_session:

            ax_user = db_session.query(AxUser).filter(
                AxUser.email == args.get('email')
            ).first()
            if ax_user:
                return CreateUser(user=None, ok=False, msg="users.email-exists")

            new_user = AxUser()
            new_user.email = args.get('email')
            new_user.name = args.get('name')
            new_user.short_name = args.get('short_name')
            new_user.info = args.get('info')
            new_user.password = pbkdf2_sha256.hash(args.get('password'))
            new_user.password_must_change = True
            new_user.is_blocked = args.get('is_blocked')
            db_session.add(new_user)
            db_session.flush()

            return CreateUser(user=new_user, ok=True, msg=None)


class UpdateUser(graphene.Mutation):
    """ Update AxUser """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        name = graphene.String()
        short_name = graphene.String()
        info = graphene.String()
        password = graphene.String(required=False, default_value=None)
        avatar_tmp = graphene.String(required=False, default_value=None)
        is_blocked = graphene.Boolean()

    ok = graphene.Boolean()
    user = graphene.Field(User)

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        # avatar_tmp = args.get('avatar_tmp')
        guid = args.get('guid')

        err = 'Error in gql mutation - users_schema -> UpdateUser.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            ax_user = db_session.query(AxUser).filter(
                AxUser.guid == uuid.UUID(guid)
            ).first()

            if args.get('name'):
                ax_user.name = args.get('name')
            if args.get('short_name'):
                ax_user.short_name = args.get('short_name')
            if args.get('info'):
                ax_user.info = args.get('info')
            if args.get('password'):
                ax_user.password = pbkdf2_sha256.hash(args.get('password'))
                ax_user.password_must_change = True

            if args.get('is_blocked') is not None:
                ax_user.is_blocked = args.get('is_blocked')

                number_of_users = db_session.query(AxUser).filter(
                    AxUser.is_group.is_(False)
                ).filter(
                    AxUser.is_blocked.is_(False)
                ).count()
                if number_of_users > ax_auth.user_num:
                    maxi = ax_auth.user_num
                    err = f"Maximum number of users exceeded! Max={maxi}"
                    logger.exception(err)
                    raise Exception(err)

            db_session.flush()
            return UpdateUser(user=ax_user, ok=True)


class ChangeUserPassword(graphene.Mutation):
    """ Update AxUser """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        password = graphene.String()

    user = graphene.Field(User)
    ok = graphene.Boolean()

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        # avatar_tmp = args.get('avatar_tmp')
        guid = args.get('guid')

        err = 'Error in gql mutation - users_schema -> ChangeUserPassword.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            ax_user = db_session.query(AxUser).filter(
                AxUser.guid == uuid.UUID(guid)
            ).first()

            current_user = info.context['user']
            user_guid = current_user.get(
                'user_id', None) if current_user else None
            if not ax_user or ax_user.guid != ax_misc.guid_or_none(user_guid):
                return None

            ax_user.password = pbkdf2_sha256.hash(args.get('password'))
            ax_user.password_must_change = False
            db_session.flush()

            return ChangeUserPassword(user=ax_user, ok=True)


class DeleteUser(graphene.Mutation):
    """ Deletes AxField witch is tab """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - form_schema -> DeleteUser.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            ax_user = db_session.query(AxUser).filter(
                AxUser.guid == uuid.UUID(guid)
            ).first()
            db_session.delete(ax_user)
            return DeleteUser(deleted=guid, ok=True)


class CreateGroup(graphene.Mutation):
    """ Creates AxUser with is_group=True """
    class Arguments:  # pylint: disable=missing-docstring
        short_name = graphene.String()

    ok = graphene.Boolean()
    user = graphene.Field(User)
    msg = graphene.String()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - users_schema -> CreateGroup.'
        with ax_model.try_catch(info.context['session'], err) as db_session:

            ax_user = db_session.query(AxUser).filter(
                AxUser.short_name == args.get('short_name')
            ).filter(
                AxUser.is_group.is_(True)
            ).first()

            if ax_user:
                return CreateGroup(
                    user=None, ok=False, msg="users.group-name-exists")

            new_user = AxUser()
            new_user.short_name = args.get('short_name')
            new_user.is_group = True
            db_session.add(new_user)
            db_session.flush()

            return CreateGroup(user=new_user, ok=True, msg=None)


class UpdateGroup(graphene.Mutation):
    """ Creates AxUser with is_group=True """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        short_name = graphene.String()

    ok = graphene.Boolean()
    user = graphene.Field(User)
    msg = graphene.String()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - users_schema -> UpdateGroup.'
        with ax_model.try_catch(info.context['session'], err) as db_session:

            existing_user = db_session.query(AxUser).filter(
                AxUser.short_name == args.get('short_name')
            ).filter(
                AxUser.is_group.is_(True)
            ).first()

            if existing_user:
                return UpdateGroup(
                    user=None, ok=False, msg="users.group-name-exists")

            ax_user = db_session.query(AxUser).filter(
                AxUser.guid == uuid.UUID(args.get('guid'))
            ).filter(
                AxUser.is_group.is_(True)
            ).first()

            ax_user.short_name = args.get('short_name')
            db_session.flush()

            return UpdateGroup(user=ax_user, ok=True, msg=None)


class AddUserToGroup(graphene.Mutation):
    """ Creates AxGroup2Users """
    class Arguments:  # pylint: disable=missing-docstring
        user_guid = graphene.String()
        group_guid = graphene.String()

    ok = graphene.Boolean()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - users_schema -> AddUserToGroup.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            user_guid = uuid.UUID(args.get('user_guid'))
            group_guid = uuid.UUID(args.get('group_guid'))

            group2user = db_session.query(AxGroup2Users).filter(
                AxGroup2Users.user_guid == user_guid
            ).filter(
                AxGroup2Users.group_guid == group_guid
            ).first()
            if group2user:
                return AddUserToGroup(ok=True)

            group2user = AxGroup2Users()
            group2user.group_guid = group_guid
            group2user.user_guid = user_guid
            db_session.add(group2user)
            db_session.flush()

            return AddUserToGroup(ok=True)


class RemoveUserFromGroup(graphene.Mutation):
    """ Deletes AxGroup2Users """
    class Arguments:  # pylint: disable=missing-docstring
        user_guid = graphene.String()
        group_guid = graphene.String()

    ok = graphene.Boolean()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - users_schema -> RemoveUserFromGroup.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            user_guid = uuid.UUID(args.get('user_guid'))
            group_guid = uuid.UUID(args.get('group_guid'))

            db_session.query(AxGroup2Users).filter(
                AxGroup2Users.user_guid == user_guid
            ).filter(
                AxGroup2Users.group_guid == group_guid
            ).delete()
            db_session.flush()

            return RemoveUserFromGroup(ok=True)


class AddUserToRole(graphene.Mutation):
    """ Creates AxRole2Users """
    class Arguments:  # pylint: disable=missing-docstring
        user_guid = graphene.String()
        role_guid = graphene.String()

    ok = graphene.Boolean()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - users_schema -> AddUserToRole.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            user_guid = uuid.UUID(args.get('user_guid'))
            role_guid = uuid.UUID(args.get('role_guid'))

            role2user = db_session.query(AxRole2Users).filter(
                AxRole2Users.user_guid == user_guid
            ).filter(
                AxRole2Users.role_guid == role_guid
            ).first()
            if role2user:
                return AddUserToRole(ok=True)

            role2user = AxRole2Users()
            role2user.role_guid = role_guid
            role2user.user_guid = user_guid
            db_session.add(role2user)
            db_session.flush()

            return AddUserToRole(ok=True)


class RemoveUserFromRole(graphene.Mutation):
    """ Removes AxRole2Users """
    class Arguments:  # pylint: disable=missing-docstring
        user_guid = graphene.String()
        role_guid = graphene.String()

    ok = graphene.Boolean()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - users_schema -> RemoveUserFromRole.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            user_guid = uuid.UUID(args.get('user_guid'))
            role_guid = uuid.UUID(args.get('role_guid'))

            db_session.query(AxRole2Users).filter(
                AxRole2Users.user_guid == user_guid
            ).filter(
                AxRole2Users.role_guid == role_guid
            ).delete()
            db_session.flush()

            return RemoveUserFromRole(ok=True)


class LogoutUser(graphene.Mutation):
    """ Delete refresh jwt token from cache """
    # class Arguments:  # pylint: disable=missing-docstring
    #     user_guid = graphene.String()
    #     role_guid = graphene.String()

    ok = graphene.Boolean()

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        """ Delete refresh token from cache """
        del args
        current_user = info.context['user']
        user_guid = current_user.get('user_id', None) if current_user else None
        if user_guid:
            key = f'refresh_token_{user_guid}'
            await ax_cache.cache.delete(key)

        return LogoutUser(ok=True)


class UsersQuery(graphene.ObjectType):
    """AxUser queryes"""
    all_users = graphene.List(
        User,
        search_string=graphene.Argument(type=graphene.String, required=False),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )
    blocked_users = graphene.List(
        User,
        search_string=graphene.Argument(type=graphene.String, required=False),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )
    all_groups = graphene.List(
        User,
        update_time=graphene.Argument(type=graphene.String, required=False)
    )
    group_users = graphene.List(
        User,
        group_guid=graphene.Argument(type=graphene.String, required=True),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )
    role_users = graphene.List(
        User,
        role_guid=graphene.Argument(type=graphene.String, required=True),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )
    page_users = graphene.List(
        User,
        page_guid=graphene.Argument(type=graphene.String, required=True),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )
    users_and_groups = graphene.List(
        User,
        search_string=graphene.Argument(type=graphene.String, required=False),
        guids=graphene.Argument(type=graphene.String, required=False),
        emails=graphene.Argument(type=graphene.String, required=False),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )
    only_users = graphene.List(
        User,
        search_string=graphene.Argument(type=graphene.String, required=False),
        guids=graphene.Argument(type=graphene.String, required=False),
        emails=graphene.Argument(type=graphene.String, required=False),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )
    find_user = graphene.Field(
        User,
        guid=graphene.Argument(type=graphene.String, required=True),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )
    current_ax_user = graphene.Field(
        User,
        update_time=graphene.Argument(type=graphene.String, required=False)
    )

    @ax_admin_only
    async def resolve_all_users(
            self, info, search_string=None, update_time=None):
        """Get all users"""
        del update_time
        err = 'Error in GQL query - all_users.'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True):
            query = User.get_query(info)  # SQLAlchemy query
            users_list = []
            if not search_string:
                users_list = query.filter(
                    AxUser.is_group.is_(False)
                ).filter(
                    AxUser.is_blocked.is_(False)
                ).all()
            else:
                users_list = query.filter(
                    AxUser.is_group.is_(False)
                ).filter(
                    AxUser.short_name.ilike(f"%{search_string}%")
                ).filter(
                    AxUser.is_blocked.is_(False)
                ).all()
            return users_list


    @ax_admin_only
    async def resolve_blocked_users(
            self, info, update_time=None):
        """Get all users"""
        del update_time
        err = 'Error in GQL query - all_users.'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True):
            query = User.get_query(info)  # SQLAlchemy query
            users_list = []
            users_list = query.filter(
                AxUser.is_group.is_(False)
            ).filter(
                AxUser.is_blocked.is_(True)
            ).all()
            return users_list


    @ax_admin_only
    async def resolve_all_groups(self, info, update_time=None):
        """Get all groups"""
        del update_time
        err = 'Error in GQL query - resolve_all_groups.'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True):
            query = User.get_query(info)  # SQLAlchemy query
            users_list = query.filter(
                AxUser.is_group.is_(True)
            ).filter(
                AxUser.is_all_users.is_(False)
            ).filter(
                AxUser.is_everyone.is_(False)
            ).filter(
                AxUser.is_blocked.is_(False)
            ).all()
            return users_list

    @ax_admin_only
    async def resolve_group_users(self, info, group_guid, update_time=None):
        """Get users of group"""
        del update_time
        err = 'Error in GQL query - resolve_all_groups.'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True):
            user_list = await ax_dialects.dialect.select_group_users(
                db_session=info.context['session'], group_guid=group_guid)
            user_guids = []
            for user in user_list:
                user_guids.append(user['guid'])
            query = User.get_query(info)  # SQLAlchemy query
            ret_list = query.filter(
                AxUser.guid.in_(user_guids)
            ).filter(
                AxUser.is_blocked.is_(False)
            ).all()
            return ret_list

    @ax_admin_only
    async def resolve_role_users(self, info, role_guid, update_time=None):
        """Get users of role"""
        del update_time
        err = 'Error in GQL query - resolve_role_users.'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True):
            user_list = await ax_dialects.dialect.select_role_users(
                db_session=info.context['session'], role_guid=role_guid)
            user_guids = []
            for user in user_list:
                user_guids.append(user['guid'])
            query = User.get_query(info)  # SQLAlchemy query
            ret_list = query.filter(
                AxUser.guid.in_(user_guids)
            ).filter(
                AxUser.is_blocked.is_(False)
            ).all()
            return ret_list

    @ax_admin_only
    async def resolve_page_users(self, info, page_guid, update_time=None):
        """Get users that allowed to view page"""
        del update_time
        err = 'Error in GQL query - resolve_page_users.'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True):
            user_list = await ax_dialects.dialect.select_page_users(
                db_session=info.context['session'], page_guid=page_guid)
            user_guids = []
            for user in user_list:
                user_guids.append(user['guid'])
            query = User.get_query(info)  # SQLAlchemy query
            ret_list = query.filter(
                AxUser.guid.in_(user_guids)
            ).filter(
                AxUser.is_blocked.is_(False)
            ).all()
            return ret_list

    async def resolve_users_and_groups(
            self,
            info,
            search_string=None,
            update_time=None,
            guids=None,
            emails=None):
        """Get all users and groups in one search"""
        del update_time, emails
        err = 'Error in GQL query - users_and_groups.'
        with ax_model.try_catch(info.context['session'], err, no_commit=True):
            query = User.get_query(info)
            users_list = []
            if search_string and guids:
                real_guids = []
                guids_dict = json.loads(guids)
                for guid in guids_dict["items"]:
                    real_guids.append(uuid.UUID(guid))

                users_list = query.filter(
                    or_(
                        AxUser.short_name.ilike(f"%{search_string}%"),
                        AxUser.guid.in_(real_guids))
                ).filter(
                    AxUser.is_blocked.is_(False)
                ).all()
            elif not search_string and guids:
                real_guids = []
                guids_dict = json.loads(guids)
                for guid in guids_dict["items"]:
                    real_guids.append(uuid.UUID(guid))

                users_list = query.filter(
                    AxUser.guid.in_(real_guids)
                ).filter(
                    AxUser.is_blocked.is_(False)
                ).all()
            elif search_string and not guids:
                users_list = query.filter(
                    AxUser.short_name.ilike(f"%{search_string}%")
                ).filter(
                    AxUser.is_blocked.is_(False)
                ).all()
            else:
                users_list = query.filter(
                    AxUser.is_blocked.is_(False)
                ).all()
            return users_list

    async def resolve_only_users(
            self, info, search_string=None, update_time=None,
            guids=None, emails=None):
        """Get all users and groups in one search"""
        del update_time
        err = 'Error in GQL query - users_and_groups.'
        with ax_model.try_catch(info.context['session'], err, no_commit=True):
            query = User.get_query(info)
            users_list = []
            if search_string and guids:
                real_guids = []
                guids_dict = json.loads(guids)
                for guid in guids_dict["items"]:
                    real_guids.append(uuid.UUID(guid))

                users_list = query.filter(
                    or_(
                        AxUser.short_name.ilike(f"%{search_string}%"),
                        AxUser.guid.in_(real_guids))
                ).filter(
                    AxUser.is_group.is_(False)
                ).filter(
                    AxUser.is_blocked.is_(False)
                ).all()
            elif search_string and emails:
                real_guids = []
                email_list = ast.literal_eval(emails)

                users_list = query.filter(
                    or_(
                        AxUser.short_name.ilike(f"%{search_string}%"),
                        AxUser.email.in_(email_list))
                ).filter(
                    AxUser.is_group.is_(False)
                ).filter(
                    AxUser.is_blocked.is_(False)
                ).all()
            elif not search_string and guids:
                real_guids = []
                guids_dict = json.loads(guids)
                for guid in guids_dict["items"]:
                    real_guids.append(uuid.UUID(guid))

                users_list = query.filter(
                    AxUser.guid.in_(real_guids)
                ).filter(
                    AxUser.is_group.is_(False)
                ).filter(
                    AxUser.is_blocked.is_(False)
                ).all()
            elif search_string and not guids:
                users_list = query.filter(
                    AxUser.short_name.ilike(f"%{search_string}%")
                ).filter(
                    AxUser.is_group.is_(False)
                ).filter(
                    AxUser.is_blocked.is_(False)
                ).all()
            elif not search_string and not guids and emails:
                email_list = ast.literal_eval(emails)
                users_list = query.filter(
                    AxUser.email.in_(email_list)
                ).filter(
                    AxUser.is_group.is_(False)
                ).filter(
                    AxUser.is_blocked.is_(False)
                ).all()

            return users_list

    @ax_admin_only
    async def resolve_find_user(self, info, guid, update_time):
        """default find method"""
        del update_time
        err = 'Error in GQL query - find_user.'
        with ax_model.try_catch(info.context['session'], err, no_commit=True):
            query = User.get_query(info)
            ax_user = query.filter(
                AxUser.guid == guid
            ).first()
            return ax_user


    async def resolve_current_ax_user(self, info, update_time):
        """ Returns current AxUser """
        del update_time
        err = 'Error in GQL query - find_user.'
        with ax_model.try_catch(info.context['session'], err, no_commit=True):
            current_user = info.context['user']
            if not current_user:
                return None

            user_guid = current_user['user_id']
            query = User.get_query(info)
            ax_user = query.filter(
                AxUser.guid == uuid.UUID(user_guid)
            ).filter(
                AxUser.is_blocked.is_(False)
            ).first()

            ax_user.is_active_admin = current_user.get('is_admin', False)
            return ax_user


class UsersSubscription(graphene.ObjectType):
    """GraphQL subscriptions"""


class UsersMutations(graphene.ObjectType):
    """Contains all AxUser mutations"""
    create_user = CreateUser.Field()
    update_user = UpdateUser.Field()
    delete_user = DeleteUser.Field()
    change_user_password = ChangeUserPassword.Field()
    create_group = CreateGroup.Field()
    update_group = UpdateGroup.Field()
    add_user_to_group = AddUserToGroup.Field()
    remove_user_from_group = RemoveUserFromGroup.Field()
    add_user_to_role = AddUserToRole.Field()
    remove_user_from_role = RemoveUserFromRole.Field()
    logout_user = LogoutUser.Field()



""" This is part of GraphQL schema (Mutaions, Queryes, Subscriptions).
Contains GQL schema for form actions - DoAction, subscribtion, avalible actions
"""
# import os
import sys
import uuid
import copy
import asyncio
# import sys
import traceback
import json
from concurrent.futures import ThreadPoolExecutor
from functools import partial
import graphene
import aiopubsub
from dotmap import DotMap
from loguru import logger

# Fields specific:
import stripe

from backend.model import AxForm, AxAction
import backend.model as ax_model
import backend.dialects as ax_dialects
import backend.pubsub as ax_pubsub
import backend.misc as ax_misc
import backend.scheduler as ax_scheduler
import backend.emails as ax_emails
import backend.exec as ax_exec
import backend.auth as ax_auth
import backend.cache as ax_cache
from backend.schemas.types import Action, Form, ConsoleMessage, \
    ActionNotifyMessage

import backend.fields.Ax1tom as AxFieldAx1tom   # pylint: disable=unused-import
import backend.fields.Ax1tomTable as AxFieldAx1tomTable  # pylint: disable=unused-import
import backend.fields.AxChangelog as AxFieldAxChangelog  # pylint: disable=unused-import
import backend.fields.AxFiles as AxFieldAxFiles  # pylint: disable=unused-import
import backend.fields.AxImageCropDb as AxFieldAxImageCropDb  # pylint: disable=unused-import
import backend.fields.AxNum as AxFieldAxNum  # pylint: disable=unused-import
import backend.fields.AxComments as AxFieldAxComments  # pylint: disable=unused-import
import backend.fields.AxApproval as AxFieldAxApproval  # pylint: disable=unused-import
import backend.fields.AxPaymentStripe as AxFieldAxPaymentStripe  # pylint: disable=unused-import
import backend.fields.AxAuthor as AxFieldAxAuthor  # pylint: disable=unused-import
import backend.fields.AxDate as AxFieldAxDate  # pylint: disable=unused-import

this = sys.modules[__name__]
action_loop = None


class InterpreterError(Exception):
    """ Class for AxAction python code errors """


class Executor:
    """In most cases, you can just use the 'execute' instance as a
    function, i.e. y = await execute(f, a, b, k=c) => run f(a, b, k=c) in
    the executor, assign result to y. The defaults can be changed, though,
    with your own instantiation of Executor, i.e. execute =
    Executor(nthreads=4)"""

    def __init__(self, loop, nthreads=1):
        self._ex = ThreadPoolExecutor(nthreads)
        self._loop = loop

    def __call__(self, func, *args, **kw):
        result = self._loop.run_in_executor(
            self._ex, partial(func, *args, **kw))
        return result


async def console_log(msg, modal_guid):
    """ Method used in AxAction terminal. It sends message to vue via ws """
    ax_pubsub.publisher.publish(
        aiopubsub.Key('console_log'),
        {
            'text': msg,
            'modal_guid': modal_guid
        })
    await asyncio.sleep(0.01)


class ConsoleSender:
    """ Used to send ax_console web socket messages to AxForm"""

    def __init__(self, modal_guid):
        self.modal_guid = modal_guid
        self.console_was_used = False

    def __call__(self, msg):
        self.console_was_used = True
        console_func = partial(console_log, msg=msg,
                               modal_guid=self.modal_guid)
        this.action_loop.call_soon_threadsafe(console_func)
        asyncio.ensure_future(console_log(msg=msg, modal_guid=self.modal_guid),
                              loop=this.action_loop)

    def check_was_used(self):
        """ If console was used once - this will return true
        Used in detecting if exception or syntex error must be outputed to ax
        console """
        return self.console_was_used


class ActionExecuter:
    """ Runs execute_action with pre-set db_session """

    def __init__(self, db_session, current_user):
        self.db_session = db_session
        self.current_user = current_user

    def __call__(
            self,
            row_guid=None,
            form_guid=None,
            form_db_name=None,
            action_guid=None,
            action_db_name=None,
            values=None,
            modal_guid=None,
            arguments=None):
        return this.execute_action(
            db_session=self.db_session,
            row_guid=row_guid,
            form_guid=form_guid,
            form_db_name=form_db_name,
            action_guid=action_guid,
            action_db_name=action_db_name,
            values=values,
            modal_guid=modal_guid,
            arguments=arguments,
            current_user=self.current_user
        )


async def get_actions(
        form,
        current_state=None,
        current_user=None,
        dynamic_role_guids=None):
    """ Get actions for current state

    Args:
        form (AxForm): Current form
        current_state (str, optional): Name of current state.
            If None, than it is create action. Defaults to None.

    Returns:
        List(AxAction): List of avalible actions for current form and state
    """
    ret_actions = []
    state_guid = None
    all_state_guid = None

    for state in form.states:
        if current_state:
            if state.name == current_state:
                state_guid = state.guid
        else:
            if state.is_start:
                state_guid = state.guid
        if state.is_all:
            all_state_guid = state.guid

    for action in form.actions:
        if action.from_state_guid == state_guid:
            ret_actions.append(action)
        if current_state and action.from_state_guid == all_state_guid:
            ret_actions.append(action)

    # Filter action based on permission cache
    perm_actions = []
    for action in ret_actions:
        action_is_avalible = await ax_auth.check_if_action_avalible(
            action=action,
            current_user=current_user,
            dynamic_role_guids=dynamic_role_guids)
        if action_is_avalible:
            perm_actions.append(action)

    return perm_actions


async def do_exec(
        db_session,
        action,
        form,
        arguments=None,
        modal_guid=None,
        current_user=None):
    """ Executes python commands form AxAction.code

    Args:
        action (AxAction): Current action that is performed
        form (AxForm): Tobe form. Its same as action.form, but it contains
            values from vue form
        arguments(**kargs): Custom dict that can be used in code
        modal_guid (str): AxForm modal guid for wss console subscribtion

    Returns:
        Dict: Its custom Dict containing:
            info (str): Message that must be displayed to user
            error (str): Error that must be displayed to user
            item (DotMap): Javascript style dictionary of field and values.
                example- item.guid -> guid of current row
            exception (Dict): Info on exception that accured
            abort (Bool): If set to True - the action will be aborted and row
                state will not be changed
    """
    import backend.schema as ax_schema

    host = await ax_misc.get_ax_host()
    localz = dict()
    ax = DotMap()  # javascript style dicts item['guid'] == item.guid
    ax.row.guid = form.row_guid
    ax.arguments = arguments
    ax.stripe = stripe
    ax.user_email = None
    ax.user_guid = None
    if current_user:
        ax.user_email = current_user.get("email", None)
        ax.user_guid = current_user.get("user_id", None)
    ax.tom_label = await ax_misc.get_tom_label(form)
    ax.host = host
    ax.form_url = f'{host}/form/{form.db_name}/{form.row_guid}'
    ax.form = form
    ax.action = action
    ax.schema = ax_schema.schema
    ax.sql = ax_dialects.dialect.custom_query
    ax.print = ConsoleSender(modal_guid=modal_guid)
    ax.email = ax_emails.email_sender
    ax.paths.uploads = ax_misc.uploads_root_dir
    ax.paths.tmp = ax_misc.tmp_root_dir
    ax.add_action_job = ax_scheduler.add_action_job
    ax.do_action = ActionExecuter(
        db_session=db_session, current_user=current_user)
    ax.modal_guid = modal_guid
    for field in form.db_fields:
        ax.row[field.db_name] = field.value
    localz['ax'] = ax
    line_number = None

    try:
        await ax_exec.aexec(code=str(action.code), localz=localz, ax=ax)
        ret_ax = localz['ax']

        ret_data = {
            "info": ret_ax.message if 'message' in ret_ax else None,
            "error": ret_ax.error if 'error' in ret_ax else None,
            "item": ret_ax.row,
            "exception": None,
            "abort": ret_ax.abort if 'abort' in ret_ax else None
        }
        return ret_data
    except SyntaxError as err:
        logger.error(err)
        error_class = err.__class__.__name__
        detail = err.args[0]
        line_number = err.lineno
        ret_data = {
            "info": None,
            "item": None,
            "error": None,
            "exception": {
                "error_class": error_class,
                "line_number": line_number,
                "action_name": action.name,
                "detail": detail
            },
            "abort": True
        }
        if ax.print.check_was_used():
            err = json.dumps(ret_data, indent=4, sort_keys=True)
            ax.print('\n\n----- SYNTAX ERROR ------\n\n' + err)
        return ret_data
    except Exception as err:    # pylint: disable=broad-except
        logger.error(err)
        error_class = err.__class__.__name__
        detail = err.args[0]
        cl, exc, tb = sys.exc_info()
        del cl, exc
        # line_number = traceback.extract_tb(tb)[-1][1]
        line_number = 0
        traces = traceback.extract_tb(tb)
        for idx, trace in enumerate(traces):
            if trace.filename == '<string>':
                line_number = traceback.extract_tb(tb)[idx].lineno
        del tb

        ret_data = {
            "info": None,
            "item": None,
            "error": None,
            "exception": {
                "error_class": error_class,
                "line_number": line_number,
                "action_name": action.name,
                "detail": detail
            },
            "abort": True
        }
        if ax.print.check_was_used():
            err = json.dumps(ret_data, indent=4, sort_keys=True)
            ax.print('\n\n----- EXCEPTION ------\n\n' + err)
        return ret_data


async def get_before_form(db_session, row_guid, form, ax_action):
    """Constructs before_form AxForm object and populate fields with
    values of specific row. AxForm.fields[0].value

    Args:
        row_guid (str): guid of specific row
        form (AxForm): empty AxForm
        ax_action (AxAction): AxAction wich is currently happening

    Returns:
       before_object (AxForm), tobe_object(AxForm): before_object is AxForm
       with field values of specific row. tobe_row is dummy for future
       population"""
    tobe_form = form

    if row_guid:
        before_result = await ax_dialects.dialect.select_one(
            db_session=db_session,
            form=tobe_form,
            fields_list=tobe_form.db_fields,
            row_guid=row_guid)

        if not before_result:
            raise Exception(
                'Error in DoAction. Cant find row for action')

        tobe_form.current_state_name = before_result[0]['axState']
        tobe_form.row_guid = before_result[0]['guid']

        if tobe_form.current_state_name != ax_action.from_state.name \
                and ax_action.from_state.is_all is False:
            raise Exception(
                'Error in DoAction. Performed action does not fit axState')

        # populate each AxField of before_form with old data
        for field in tobe_form.db_fields:
            if field.db_name in before_result[0].keys():
                field.value = before_result[0][field.db_name]
                if (field.value and isinstance(field.value, str)
                        and field.field_type.value_type == 'JSON'):
                    field.value = json.loads(field.value)

    before_form = copy.deepcopy(tobe_form)
    return before_form, tobe_form


async def run_field_backend(db_session, field, action, before_form, tobe_form,
                            current_user, when='before', query_type='update'):
    """Some field types need backend execution on insert/update/delete

    Args:
        field (AxField): contains current .value of updated field
        action (AxAction): Just for info
        before_form (AxForm): Just for info
        tobe_form (AxForm): Just for info
        current_user (AxUser): Just for info
        when (str, optional): Can be 'before' or 'after'.
        query_type (str, optional): Can be 'insert', 'update', 'delete'.
        Defaults to 'update'.

    Returns:
        object: Returns new field value
    """
    tag = field.field_type.tag
    # field_py_file_path = f"backend/fields/{tag}.py"
    function_name = f"{when}_{query_type}"

    # if os.path.exists(ax_misc.path(field_py_file_path)):
    # field_py = globals()[f'AxField{tag}']
    field_py = globals().get(f'AxField{tag}', None)
    if field_py and hasattr(field_py, function_name):
        method_to_call = getattr(field_py, function_name)
        new_value = await method_to_call(
            db_session=db_session,
            field=field,
            before_form=before_form,
            tobe_form=tobe_form,
            action=action,
            current_user=current_user)
        return new_value
    return field.value


async def execute_action(
        db_session,
        row_guid=None,
        form_guid=None,
        form_db_name=None,
        action_guid=None,
        action_db_name=None,
        values=None,
        modal_guid=None,
        arguments=None,
        current_user=None):
    """ Performs Action on row
        # 0. Get AxForm
        # 1. Get AxAction and query_type
        # 2. Check if user can perform current action
        # 3. Get before_form - fill it with values from database row
        # 4. Assemble tobe_object - fill it with values from AxForm.vue
        # 5. Run before backend code for each field.
            FieldType can have backend code wich is performed before and after
            action. Sea backend.fields for more info.
        # 6. Run python action, rewrite tobe_object
        # 7. Check fields permissions.
        # 8. Make update or insert or delete query. Db commit is here.
        # 9. Run after backend code for each field.
        # 10. Fire all web-socket subscribtions. Notify of action performed
            AxForm.vue will display message, that current row have been modified
            AxGrid.vue will reload data of grid

    Args:
        row_guid (str): Guid of row that is opened (null if new row)
        form_guid (str): Guid of AxForm that is opened
        form_db_name (str): db_name of AxForm that is opened.
            (Form guid or db_name must be specified)
        action_guid (str): Guid of AxAction that is performed
        action_db_name (str): db_name of AxAction
            (Action guid or db_name must be specified)
        values (Dict): Dict from AxForm.vue with values of form
        modal_guid (str): Guid generated by AxForm.vue. It used in web-socket
            subscribtion. If current form is performing this action - it does
            not need to notify user of performed action

    Returns:
        form (AxForm): Form that is changed by action. The field values could
            have been changed, if row is created - the row_guid is changed,
            state of current row changed.
        new_guid (str): Guid of current row. Null if row deleted
        messages (Dict): Custom dict containing messages, errors, exceptions
            from executing python code of action. Sea do_exec func for info.
        modal_guid (str): Same as Args. It used in web-socket
            subscribtion. If current form is performing this action - it does
            not need to notify user of performed action
        ok (bool): Used in gql query
    """
    try:
        err = 'action_schema -> execute_action'
        with ax_model.try_catch(db_session, err) as db_session:
            new_guid = uuid.uuid4()
            if not values:
                values = {}

            # Check if row locked, if not -> lock it
            if row_guid:
                lock_key = f'lock_{row_guid}'
                is_locked = await ax_cache.cache.get(lock_key)
                if not is_locked:
                    await ax_cache.cache.set(lock_key, True)
                else:
                    messages = {
                        "exception": None,
                        "error": "$i18n('form.row-locked')",
                        "info": None
                    }
                    messages_json = json.dumps(messages)
                    return {
                        "form": None,
                        "new_guid": None,
                        "messages": messages_json,
                        "modal_guid": modal_guid,
                        "ok": False
                    }

            # 0. Get AxForm
            ax_form = None
            if form_guid:
                ax_form = db_session.query(AxForm).filter(
                    AxForm.guid == uuid.UUID(form_guid)
                ).first()
            elif form_db_name:
                ax_form = db_session.query(AxForm).filter(
                    AxForm.db_name == form_db_name
                ).first()
            else:
                raise ValueError(
                    'doAction GQL - form guid or db_name required')

            db_session.expire(ax_form)

            # 1. Get AxAction and query_type
            ax_action = None
            if not action_guid and not action_db_name:
                raise ValueError(
                    'doAction GQL - action guid or db_name required')
            for action in ax_form.actions:
                if not ax_action:
                    if action_guid and action.guid == uuid.UUID(action_guid):
                        ax_action = action
                    if action_db_name and action.db_name == action_db_name:
                        ax_action = action

            query_type = 'update'
            if ax_action.from_state.is_start:
                query_type = 'insert'
            elif ax_action.to_state.is_deleted:
                query_type = 'delete'

            # 3. Get before_form - fill it with values from database row
            before_form, tobe_form = await get_before_form(
                db_session=db_session,
                row_guid=row_guid,
                form=ax_form,
                ax_action=ax_action)

            # 4. Assemble tobe_object - fill it with values from AxForm.vue
            if not tobe_form.row_guid:
                tobe_form.row_guid = new_guid
            for idx, field in enumerate(tobe_form.fields):
                if field.is_tab is False and not field.field_type.is_virtual:
                    if field.db_name in values.keys():
                        tobe_form.fields[idx].value = values[field.db_name]
                        tobe_form.fields[idx].needs_sql_update = True
                    if field.field_type.is_updated_always:
                        tobe_form.fields[idx].needs_sql_update = True

            # 2. Check if user can perform current action
            dynamic_role_guids = await ax_auth.get_dynamic_roles_guids(
                ax_form=tobe_form,
                current_user=current_user)

            action_is_avalible = await ax_auth.check_if_action_avalible(
                action=ax_action,
                current_user=current_user,
                dynamic_role_guids=dynamic_role_guids)
            if not action_is_avalible:
                email = 'Anon'
                if current_user:
                    email = current_user.get('email', None)
                msg = (
                    f'Error in DoAction. Action [{ax_action.name}]',
                    f'not allowed for user [{email}]'
                )
                logger.error(msg)
                raise Exception(msg)

            # 5. Run before backend code for each field.
            for idx, field in enumerate(tobe_form.fields):
                if (field.is_tab is False and
                        field.field_type.is_backend_available):
                    tobe_form.fields[idx].value = await run_field_backend(
                        db_session=db_session,
                        when='before',
                        query_type=query_type,
                        field=field,
                        action=ax_action,
                        before_form=before_form,
                        tobe_form=tobe_form,
                        current_user=current_user)

            # 6. Run python action, rewrite tobe_object
            messages = None
            messages_json = None
            if ax_action.code:
                code_result = await do_exec(
                    db_session=db_session,
                    action=ax_action,
                    form=tobe_form,
                    arguments=arguments,
                    modal_guid=modal_guid,
                    current_user=current_user)
                messages = {
                    "exception": code_result["exception"],
                    "error": code_result["error"],
                    "info": code_result["info"]
                }
                messages_json = json.dumps(messages)
                if code_result['abort'] or code_result["exception"]:
                    await ax_cache.cache.delete(f'lock_{row_guid}')
                    return {
                        "form": before_form,
                        "new_guid": None,
                        "messages": messages_json,
                        "modal_guid": modal_guid,
                        "ok": False
                    }
                # update fields with values recieved from actions python
                new_item = code_result['item']
                for field in tobe_form.db_fields:
                    if field.db_name in new_item.keys():
                        if field.value != new_item[field.db_name]:
                            field.value = new_item[field.db_name]
                            field.needs_sql_update = True
                            field.force_sql_update = True

            # 7. Check fields permissions. If not permited - set needs_update
            # to False
            ax_form = await ax_auth.set_form_visibility(
                ax_form=ax_form,
                state_guid=ax_action.from_state.guid,
                current_user=current_user,
                dynamic_role_guids=dynamic_role_guids)

            # 8. Make update or insert or delete query #COMMIT HERE
            return_guid = tobe_form.row_guid
            after_form = copy.deepcopy(tobe_form)
            if query_type == 'insert':
                await ax_dialects.dialect.insert(
                    db_session=db_session,
                    form=tobe_form,
                    to_state_name=ax_action.to_state.name,
                    new_guid=new_guid
                )
                tobe_form.row_guid = new_guid
                return_guid = new_guid
            elif query_type == 'delete':
                await ax_dialects.dialect.delete(
                    db_session=db_session,
                    form=tobe_form,
                    row_guid=tobe_form.row_guid
                )
                return_guid = None
            else:
                await ax_dialects.dialect.update(
                    db_session=db_session,
                    form=tobe_form,
                    to_state_name=ax_action.to_state.name,
                    row_guid=tobe_form.row_guid
                )

            # 9. Run after backend code for each field.
            for field in after_form.no_tab_fields:
                if field.field_type.is_backend_available:
                    field.value = await run_field_backend(
                        db_session=db_session,
                        when='after',
                        query_type=query_type,
                        field=field,
                        action=ax_action,
                        before_form=before_form,
                        tobe_form=after_form,
                        current_user=current_user)

            db_session.commit()

            # 10. Fire all web-socket subscribtions. Notify of action performed
            notify_message = {
                "form_guid": tobe_form.guid,
                "form_icon": tobe_form.icon,
                "form_db_name": tobe_form.db_name,
                "row_guid": tobe_form.row_guid,
                "modal_guid": modal_guid,
                "action_guid": ax_action.guid,
                "action_db_name": ax_action.db_name,
                "action_icon": ax_action.icon
            }

            ax_pubsub.publisher.publish(
                aiopubsub.Key('do_action'), notify_message)

            await ax_cache.cache.delete(f'lock_{row_guid}')
            await console_log('ax_console::reload', modal_guid)
            return {
                "form": tobe_form,
                "new_guid": return_guid,
                "messages": messages_json,
                "modal_guid": modal_guid,
                "ok": True
            }
    except Exception:
        await ax_cache.cache.delete(f'lock_{row_guid}')
        logger.exception('Error while executing action - DoAction.')
        raise


def run_async(corofn, *args):
    """  run_in_executor takes only sync function, but with
    this hack we can run corutines """
    loop = asyncio.new_event_loop()
    coro = corofn(*args)
    asyncio.set_event_loop(loop)
    return loop.run_until_complete(coro)


async def run_execute_action(db_session,
                             row_guid=None,
                             form_guid=None,
                             form_db_name=None,
                             action_guid=None,
                             action_db_name=None,
                             values=None,
                             modal_guid=None,
                             arguments=None,
                             current_user=None
                             ):
    """ This function executes execute_action corutine in different thread """
    loop = asyncio.get_event_loop()
    executor = ThreadPoolExecutor(max_workers=5)
    execute_func = partial(
        execute_action,
        db_session=db_session,
        row_guid=row_guid,
        form_guid=form_guid,
        form_db_name=form_db_name,
        action_guid=action_guid,
        action_db_name=action_db_name,
        values=values,
        modal_guid=modal_guid,
        arguments=arguments,
        current_user=current_user
    )

    result = await loop.run_in_executor(
        executor,
        run_async,
        execute_func)

    return result


class DoAction(graphene.Mutation):
    """
        Runs execute_action. Performs action on row.
    """
    class Arguments:  # pylint: disable=missing-docstring
        row_guid = graphene.String(required=False, default_value=None)
        form_guid = graphene.String(required=False, default_value=None)
        form_db_name = graphene.String(required=False, default_value=None)
        action_guid = graphene.String(required=False, default_value=None)
        action_db_name = graphene.String(required=False, default_value=None)
        values = graphene.String()
        arguments = graphene.String(required=False, default_value=None)
        modal_guid = graphene.String(required=False, default_value=None)

    form = graphene.Field(Form)
    new_guid = graphene.String()
    messages = graphene.String()
    modal_guid = graphene.String()
    ok = graphene.Boolean()

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - action_schema -> DoAction.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            current_user = info.context['user']
            values_string = args.get('values')
            values = {}
            if values_string and values_string != 'null':
                values = json.loads(values_string)
            row_guid = args.get('row_guid')
            modal_guid = args.get('modal_guid')
            form_guid = args.get('form_guid')
            form_db_name = args.get('form_db_name')
            action_guid = args.get('action_guid')
            action_db_name = args.get('action_db_name')
            arguments_string = args.get('arguments')
            arguments = {}
            if arguments_string and arguments_string != 'null':
                arguments = json.loads(arguments_string)

            result = await run_execute_action(
                db_session=db_session,
                row_guid=row_guid,
                form_guid=form_guid,
                form_db_name=form_db_name,
                action_guid=action_guid,
                action_db_name=action_db_name,
                values=values,
                arguments=arguments,
                modal_guid=modal_guid,
                current_user=current_user)

            return DoAction(
                form=result["form"],
                new_guid=result["new_guid"],
                messages=result["messages"],
                modal_guid=result["modal_guid"],
                ok=result["ok"])


class ActionSubscription(graphene.ObjectType):
    """GraphQL subscriptions"""

    action_notify = graphene.Field(
        ActionNotifyMessage,
        form_db_name=graphene.Argument(type=graphene.String, required=True),
        row_guid=graphene.Argument(type=graphene.String, required=False))

    console_notify = graphene.Field(
        ConsoleMessage,
        modal_guid=graphene.Argument(type=graphene.String, required=True))

    async def resolve_action_notify(self, info, form_db_name, row_guid=None):
        """ Web-socket subscription on every performed action
            AxForm.vue will display message, that current row have been modified
            AxGrid.vue will reload data of grid

            Args:
                form_db_name (str): db_name of AxForm
                row_guid (str): If it is set, then the subscriber will be
                    notified only on actions performed with surtain row.
                    AxGrid is notified on every action
                    AxForm is notified only on row actions

            Returns:
                payload (Dict): Dict containing action info:
                    form_guid (str): AxForm.guid of current form
                    form_icon (str): font-awesome icon
                    form_db_name (str): AxForm.db_name of current form
                    row_guid (str): Guid of row on wich action is performed
                    modal_guid (str): Guid generated by AxForm.vue.
                        It used in web-socket subscribtion. If current form is
                        performing this action - it doesnot need to notify user
                        of performed action
                    action_guid (str): Current action guid
                    action_db_name (str): Db name of current action
                    action_icon (str): font-awesome icon of current action

        """
        try:
            del info
            subscriber = aiopubsub.Subscriber(
                ax_pubsub.hub, 'action_notify_subscriber')
            subscriber.subscribe(aiopubsub.Key('do_action'))
            while True:
                key, payload = await subscriber.consume()
                del key
                if payload['form_db_name'] == form_db_name:
                    if row_guid is None or row_guid == payload['row_guid']:
                        message = ActionNotifyMessage(
                            form_guid=payload['form_guid'],
                            form_icon=payload['form_icon'],
                            form_db_name=payload['form_db_name'],
                            row_guid=payload['row_guid'],
                            modal_guid=payload['modal_guid'],
                            action_guid=payload['action_guid'],
                            action_db_name=payload['action_db_name'],
                            action_icon=payload['action_icon'])
                        yield message
        except asyncio.CancelledError:
            await subscriber.remove_all_listeners()

    async def resolve_console_notify(self, info, modal_guid):
        """ Web-socket subscription on ax terminal print
        """
        try:
            del info
            subscriber = aiopubsub.Subscriber(
                ax_pubsub.hub, 'action_notify_subscriber')
            subscriber.subscribe(aiopubsub.Key('console_log'))
            this.action_loop = asyncio.get_event_loop()
            while True:
                key, payload = await subscriber.consume()
                await asyncio.sleep(0.1)
                del key
                if payload['modal_guid'] == modal_guid:
                    message = ConsoleMessage(
                        text=payload['text'],
                        modal_guid=payload['modal_guid'])
                    yield message
        except asyncio.CancelledError:
            await subscriber.remove_all_listeners()
        except Exception:  # pylint: disable=broad-except
            logger.exception('Error in gql sub resolve_console_notify.')


class ActionQuery(graphene.ObjectType):
    """Query all data of AxAction and related classes"""
    action_data = graphene.Field(
        Action,
        guid=graphene.Argument(type=graphene.String, required=True),
        update_time=graphene.Argument(type=graphene.String, required=False))
    actions_avalible = graphene.List(
        Action,
        form_db_name=graphene.Argument(type=graphene.String, required=True),
        current_state=graphene.Argument(type=graphene.String, required=False),
        update_time=graphene.Argument(type=graphene.String, required=False))

    async def resolve_action_data(self, info, guid=None, update_time=None):
        """ Gets AxAction by guid. Used in TheActionModal.vue. """
        del update_time
        err = 'action_schema -> resolve_action_data'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True):
            query = Action.get_query(info=info)
            ax_action = query.filter(AxAction.guid == uuid.UUID(guid)).first()
            return ax_action

    async def resolve_actions_avalible(self, info, form_db_name,
                                       current_state=None, update_time=None):
        """ Gets AxActions for current state """
        current_user = info.context['user']
        del update_time
        err = (f"action_schema.resolve_actions_avalible."
               f"Error in gql query - resolve_actions_avalible.")
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True) as db_session:
            ax_form = db_session.query(AxForm).filter(
                AxForm.db_name == form_db_name
            ).first()
            ax_actions = None
            if ax_form:
                ax_actions = await get_actions(
                    form=ax_form,
                    current_state=current_state,
                    current_user=current_user)
            return ax_actions


class ActionMutations(graphene.ObjectType):
    """Combine all mutations"""
    do_action = DoAction.Field()



""" Sends emails
Use like this -
    import backend.emails as ax_emails
    ax_emails.email_sender(to='', text='', html=None, subject='')

"""

import sys
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.header import Header
from loguru import logger
import aiosmtplib

this = sys.modules[__name__]
email_sender = None


def init_email(smtp_host, smtp_login, smtp_password, port=465, ssl=False):
    """ create EmailSender, remember creds """
    this.email_sender = EmailSender(
        smtp_host=smtp_host,
        smtp_login=smtp_login,
        smtp_password=smtp_password,
        port=port,
        ssl=ssl
    )


class EmailSender:
    """ Used to send emails """

    def __init__(self, smtp_host, smtp_login, smtp_password, port=465,
                 ssl=False):
        self.smtp_host = smtp_host
        self.smtp_login = smtp_login
        self.smtp_password = smtp_password
        self.port = port
        self.ssl = ssl

    async def __call__(self,    # pylint: disable=invalid-name
                       to,
                       text='Hello world',
                       html=None,
                       subject='Message from AX'):
        try:
            recipients = []
            if isinstance(to, list):
                recipients = to
            else:
                recipients.append(to)

            msg = MIMEMultipart('alternative')

            msg['Subject'] = Header(subject, "utf-8").encode()
            msg['From'] = self.smtp_login
            msg['To'] = ", ".join(recipients)

            if not html:
                html = text.replace('\n', '<br>')
            part1 = MIMEText(text.encode('utf-8'), 'plain', 'utf-8')
            part2 = MIMEText(html.encode('utf-8'), 'html', 'utf-8')

            msg.attach(part1)
            msg.attach(part2)

            await aiosmtplib.send(
                msg,
                recipients=recipients,
                sender=self.smtp_login,
                hostname=self.smtp_host,
                port=self.port,
                username=self.smtp_login,
                password=self.smtp_password,
                use_tls=self.ssl
            )
        except Exception:
            logger.exception('Error sending email with Ax email sender')
            raise



"""added options_json to Action, Role, Form

Revision ID: b3eb6ccd4b1f
Revises: 9b898740929f
Create Date: 2020-04-19 16:43:56.927231

"""
import sys
from alembic import op
import sqlalchemy as sa
from pathlib import Path

root_path = Path(__file__).resolve().parent.parent.parent.parent
sys.path.insert(0, str(root_path))

import backend.model


# revision identifiers, used by Alembic.
revision = 'b3eb6ccd4b1f'
down_revision = '9b898740929f'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('_ax_actions', sa.Column('private_options_json', sa.JSON))


def downgrade():
    op.drop_column('_ax_actions', 'private_options_json')



"""AxFiles field type functions - before, after / insert, update, delete"""
import os
import uuid
import shutil
import backend.misc as ax_misc
from backend.model import AxMessageThread, AxMessage


async def before_insert(db_session, field, before_form, tobe_form, action,
                        current_user):
    """ Create AxThread  """
    thread_guid = uuid.uuid4()
    new_thread = AxMessageThread()
    new_thread.guid = thread_guid
    new_thread.field_guid = field.guid
    new_thread.row_guid = tobe_form.row_guid
    db_session.add(new_thread)
    db_session.flush()

    field.needs_sql_update = True
    field.value = thread_guid
    return field.value


async def after_delete(db_session, field, before_form, tobe_form, action,
                       current_user):
    """ Delete related AxMessageThread. In cascade it deletes all related
        AxMessages """
    thread = db_session.query(AxMessageThread).filter(
        AxMessageThread.guid == uuid.UUID(field.value)
    ).first()
    if thread:
        db_session.query(AxMessage).filter(
            AxMessage.thread_guid == thread.guid
        ).delete()
        db_session.flush()

        db_session.delete(thread)
        db_session.commit()

    return field.value



""" GQL queries used in fields. For now only AxMessages, AxApproval"""

import uuid
import json
import asyncio
import aiopubsub
import graphene
from loguru import logger
from sqlalchemy.orm import joinedload
import stripe

from backend.auth import ax_admin_only
from backend.model import AxMessage, AxMessageThread, AxForm, AxField
import backend.model as ax_model
import backend.misc as ax_misc
import backend.auth as ax_auth
import backend.pubsub as ax_pubsub

# import backend.emails as ax_emails
# import backend.cache as ax_cache
import backend.dialects as ax_dialects

from backend.schemas.types import Message, Field, RowInfo
# from backend.auth import ax_admin_only


class CreateMessage(graphene.Mutation):
    """ Creates AxMessage """
    class Arguments:  # pylint: disable=missing-docstring
        thread_guid = graphene.String()
        text = graphene.String()
        data_json = graphene.String(required=False, default_value=None)
        parent = graphene.String(required=False, default_value=None)

    ok = graphene.Boolean()
    message = graphene.Field(Message)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - fields_schema -> CreateMessage.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            thread_guid = args.get('thread_guid')
            text = args.get('text')
            data_json = args.get('data_json')
            current_user = info.context['user']
            user_guid = current_user.get('user_id', None)
            user_email = current_user.get('email', None)
            parent = args.get('parent')

            data = None
            if data_json:
                try:
                    data = json.loads(data_json)
                except ValueError:
                    logger.exception(
                        'Error decoding data_json on creating AxMessage')

            thread = db_session.query(AxMessageThread).filter(
                AxMessageThread.guid == uuid.UUID(args.get('thread_guid'))
            ).first()
            if not thread:
                thread = AxMessageThread()
                thread.guid = uuid.UUID(thread_guid)
                db_session.add(thread)
                db_session.flush()

            new_message = AxMessage()
            new_message.author_guid = ax_misc.guid_or_none(user_guid)
            new_message.author_email = user_email
            new_message.text = text
            new_message.data_json = data
            new_message.thread_guid = thread.guid
            new_message.parent = ax_misc.guid_or_none(parent)

            db_session.add(new_message)
            db_session.flush()

            created_message = db_session.query(AxMessage).filter(
                AxMessage.guid == new_message.guid
            ).options(joinedload(AxMessage.author)).first()

            # joinedload(Event.user)

            db_session.expunge(created_message)
            db_session.expunge(created_message.author)
            ax_pubsub.publisher.publish(
                aiopubsub.Key('thread_message'), {
                    "thread_guid": str(created_message.thread_guid),
                    "ax_message": created_message
                })

            return CreateMessage(message=new_message, ok=True)


class GetStripeIntent(graphene.Mutation):
    """ Gets intent data from Stripe. Used to refresh payment status """
    class Arguments:  # pylint: disable=missing-docstring
        inten_id = graphene.String()

    ok = graphene.Boolean()
    intent = graphene.String()

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        try:
            del info
            inten_id = args.get('inten_id')
            intent = stripe.PaymentIntent.retrieve(id=inten_id)
            intent_str = json.dumps(intent)

            ok = True
            return GetStripeIntent(intent=intent_str, ok=ok)
        except Exception:
            logger.exception('Error in gql mutation - GetStripeIntent.')
            raise


class FieldsSubscription(graphene.ObjectType):
    """GraphQL subscriptions"""

    thread_notify = graphene.Field(
        Message,
        thread_guid=graphene.Argument(type=graphene.String, required=True))

    async def resolve_thread_notify(self, info, thread_guid):
        """ Web-socket subscription on new messages in thread """
        try:
            del info
            subscriber = aiopubsub.Subscriber(
                ax_pubsub.hub, 'message_notify_subscriber')
            subscriber.subscribe(aiopubsub.Key('thread_message'))
            while True:
                key, payload = await subscriber.consume()
                await asyncio.sleep(0.1)
                del key
                if payload['thread_guid'] == thread_guid:
                    ax_message = payload['ax_message']
                    yield ax_message
        except asyncio.CancelledError:
            await subscriber.remove_all_listeners()
        except Exception:  # pylint: disable=broad-except
            logger.exception('Error in gql sub resolve_thread_notify.')


class FieldsQuery(graphene.ObjectType):
    """AxFields queryes"""
    thread_messages = graphene.List(
        Message,
        thread_guid=graphene.Argument(type=graphene.String),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )
    to1_fields = graphene.List(
        Field,
        child_form_guid=graphene.Argument(type=graphene.String),
        field_guid=graphene.Argument(type=graphene.String),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )

    to1_references = graphene.List(
        RowInfo,
        row_guid=graphene.Argument(type=graphene.String),
        field_guid=graphene.Argument(type=graphene.String),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )

    async def resolve_thread_messages(
            self, info, thread_guid, update_time=None):
        """Get all AxMessages for specific thread. """
        del update_time
        err = 'Error in GQL query - resolve_thread_messages.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            current_user = info.context['user']

            if not thread_guid:
                return None

            # get thead
            thread = db_session.query(AxMessageThread).filter(
                AxMessageThread.guid == uuid.UUID(thread_guid)
            ).first()

            if not thread:
                return None

            # check field row permission
            perm = True
            if thread.field_guid and thread.row_guid:
                perm = await ax_auth.check_field_perm(
                    db_session=db_session,
                    current_user=current_user,
                    field_guid=thread.field_guid,
                    row_guid=thread.row_guid)

            if not perm:
                return None

            # get messages
            messages = thread.messages
            return messages

    @ax_admin_only
    async def resolve_to1_fields(
            self, info, child_form_guid, field_guid, update_time=None):
        """Get all AxMessages for specific thread. """
        del update_time
        err = 'Error in GQL query - resolve_to1_fields.'
        with ax_model.try_catch(info.context['session'], err) as db_session:

            ret_fields = []
            child_form = db_session.query(AxForm).filter(
                AxForm.guid == uuid.UUID(str(child_form_guid))
            ).first()

            current_field = db_session.query(AxField).filter(
                AxField.guid == uuid.UUID(str(field_guid))
            ).first()
            parent_form_db_name = str(current_field.form.db_name)

            for field in child_form.fields:
                if field.is_to1_field:
                    options = field.options
                    if options and options['form'] == parent_form_db_name:
                        db_session.expunge(field)
                        ret_fields.append(field)

            return ret_fields

    async def resolve_to1_references(
            self, info, field_guid, row_guid, update_time=None):
        """ Used in Ax1to1Children. Selects all 1to1 references to parent row
        """
        del update_time
        err = 'Error in GQL query - resolve_to1_fields.'
        with ax_model.try_catch(info.context['session'], err) as db_session:

            # Check if user have access to field

            ret_guids = []
            parent_field = db_session.query(AxField).filter(
                AxField.guid == uuid.UUID(str(field_guid))
            ).first()

            if parent_field.options and parent_field.options['field']:
                child_form_db_name = parent_field.options['form']

                child_field_guid = parent_field.options['field']
                child_field = db_session.query(AxField).filter(
                    AxField.guid == uuid.UUID(str(child_field_guid))
                ).first()

            result = await ax_dialects.dialect.select_to1_children(
                db_session=db_session,
                child_form_db_name=child_form_db_name,
                child_field_db_name=child_field.db_name,
                parent_row_guid=row_guid)

            if result:
                for row in result:
                    ret_guids.append(RowInfo(guid=row['guid']))

            return ret_guids


class FieldsMutations(graphene.ObjectType):
    """Contains all mutations"""
    create_message = CreateMessage.Field()
    get_stripe_intent = GetStripeIntent.Field()



"""
Module for database migration using Alembic
It checks if database is accesible.
Creates sqlite database if needed.
Checks if database have SqlAlchemy tables. If not - it creates tables.
    When tables are created, they are filled with default values.
    Such as AxFieldType are filled with field types.
Checks that current Database schema fits SqlAlchemy schema. If it is not,
    it tryes to upgrade it by running alembic migration scripts.

"""
import os
import sys
import uuid
import asyncio
import json
from loguru import logger
import yaml
from passlib.hash import pbkdf2_sha256
from alembic.config import Config
from alembic.migration import MigrationContext
from alembic.autogenerate import compare_metadata
from alembic import command
import backend.model as ax_model
import backend.misc as ax_misc
import backend.auth as ax_auth
from backend.model import AxAlembicVersion, AxFieldType, AxUser, AxGroup2Users,\
    AxPage, AxPage2Users, AxForm, AxAction, AxGrid, AxRole

this = sys.modules[__name__]
alembic_cfg = None
stats = None


def init_alembic_config() -> None:
    """Create Alembic config"""
    try:
        ini_path = ax_misc.path("alembic.ini")
        this.alembic_cfg = Config(ini_path)
        alembic_folder = ax_misc.path("backend/alembic")
        this.alembic_cfg.set_main_option(
            'script_location', str(alembic_folder))
    except Exception:
        logger.exception('Error initating alembic config.')
        raise


def tables_exist() -> bool:
    """Checks if database tables are created"""
    try:
        tables_created = ax_model.engine.dialect.has_table(
            ax_model.engine, '_ax_alembic_version')
        if tables_created:
            return True
        else:
            return False
    except Exception:
        logger.exception('Error in detecting if database tables are created')
        raise


def create_tables() -> None:
    """Create database and create baseline version in Alembic"""
    with ax_model.scoped_session("migration -> create_tables") as db_session:
        if os.environ.get('AX_DB_REVISION') is None:
            msg = 'Cant find AX_DB_REVISION in enviroment variables or app.yaml'
            logger.error(msg)
            raise Exception(msg)

        # ax_model.Base.query = db_session.query_property()
        ax_model.Base.metadata.create_all(ax_model.engine)
        first_version = AxAlembicVersion()
        first_version.version_num = os.environ.get('AX_DB_REVISION')
        db_session.add(first_version)

        logger.info('Ax tables not found. Creating database tables.')
        return True


def create_default_pages():
    """ Create default page """
    with ax_model.scoped_session(
            "migration -> create_default_pages") as db_session:
        index_page = AxPage()
        index_page.name = "Index Page"
        index_page.code = (f"<h1>Welcome to Ax pages</h1>\n"
                           f"<br/><br/>Hello world")
        db_session.add(index_page)

        all_user = db_session.query(AxUser).filter(
            AxUser.is_all_users.is_(True)
        ).first()

        p2u = AxPage2Users()
        p2u.page_guid = index_page.guid
        p2u.user_guid = all_user.guid
        db_session.add(p2u)

        db_session.commit()


def create_default_users():
    """ Create Admin group, Default admin, Everyone, All users """
    msg = "Error -> Migration -> create_default_users"
    with ax_model.scoped_session(msg) as db_session:
        # Admin group
        admin_group = AxUser()
        # admin_group.short_name = 'users.admin-group-name'
        admin_group.short_name = 'Administrators'
        admin_group.is_group = True
        admin_group.is_admin = True
        db_session.add(admin_group)

        # Default admin
        admin = AxUser()
        admin.email = 'default@ax-workflow.com'
        admin.password = pbkdf2_sha256.hash('deleteme')
        admin.name = 'Default Administrator. Delete this user.'
        admin.short_name = 'Default Admin'
        db_session.add(admin)

        db_session.commit()

        group_user = AxGroup2Users()
        group_user.group_guid = admin_group.guid
        group_user.user_guid = admin.guid
        db_session.add(group_user)

        # Everyone
        everyone_group = AxUser()
        # everyone_group.short_name = 'users.everyone-group-name'
        everyone_group.short_name = 'Everyone'
        everyone_group.is_group = True
        everyone_group.is_everyone = True
        db_session.add(everyone_group)

        # All users
        all_group = AxUser()
        # all_group.short_name = 'users.all-users-group-name'
        all_group.short_name = 'All users'
        all_group.is_group = True
        all_group.is_all_users = True
        db_session.add(all_group)
        db_session.commit()


def create_field_type(db_session, item):
    """ Create AxFieldType from yaml data """
    db_session.add(AxFieldType(
        tag=item.get('tag', None),
        name=item.get('name', None),
        position=item.get('position', 0),
        default_name=item.get('default_name', None),
        default_db_name=item.get('default_db_name', ''),
        value_type=item.get('value_type', ''),
        parent=item.get('parent', ''),
        icon=item.get('icon', ''),
        comparator=item.get('comparator', ''),
        virtual_source=item.get('virtual_source', None),
        is_virtual=item.get('is_virtual', False),
        is_readonly=item.get('is_readonly', False),
        is_inline_editable=item.get('is_inline_editable', False),
        is_backend_available=item.get('is_backend_available', False),
        is_display_backend_avalible=item.get(
            'is_display_backend_avalible', False),
        is_columnn_avalible=item.get('is_columnn_avalible', False),
        is_updated_always=item.get('is_updated_always', False),
        is_group=item.get('is_group', False),
        is_always_whole_row=item.get('is_always_whole_row', False)
    ))


def update_field_type(field_types, item):
    """ Update AxFieldType from yaml data if needed """
    for ax_field in field_types:
        if ax_field.tag == item['tag']:
            # TODO do this from list in loop!
            if 'name' in item and ax_field.name != item['name']:
                ax_field.name = item['name']
            if 'position' in item and ax_field.position != item['position']:
                ax_field.position = item['position']
            if ('default_name' in item and
                    ax_field.default_name != item['default_name']):
                ax_field.name = item['default_name']
            if ('default_db_name' in item and
                    ax_field.default_db_name != item['default_db_name']):
                ax_field.default_db_name = item['default_db_name']
            if ('value_type' in item and
                    ax_field.value_type != item['value_type']):
                ax_field.value_type = item['value_type']
            if 'parent' in item and ax_field.parent != item['parent']:
                ax_field.parent = item['parent']
            if 'icon' in item and ax_field.icon != item['icon']:
                ax_field.icon = item['icon']
            if ('comparator' in item and
                    ax_field.comparator != item['comparator']):
                ax_field.comparator = item['comparator']
            if ('virtual_source' in item
                    and ax_field.virtual_source != item['virtual_source']):
                ax_field.virtual_source = item['virtual_source']
            if ('is_virtual' in item
                    and ax_field.is_virtual != item['is_virtual']):
                ax_field.is_virtual = item['is_virtual']
            if ('is_readonly' in item
                    and ax_field.is_readonly != item['is_readonly']):
                ax_field.is_readonly = item['is_readonly']
            if ('is_inline_editable' in item and
                    ax_field.is_inline_editable !=
                    item['is_inline_editable']):
                ax_field.is_inline_editable = item['is_inline_editable']
            if ('is_backend_available' in item and
                    ax_field.is_backend_available !=
                    item['is_backend_available']):
                ax_field.is_backend_available = item['is_backend_available']
            if ('is_display_backend_avalible' in item and
                    ax_field.is_display_backend_avalible !=
                    item['is_display_backend_avalible']):
                the_item = item['is_display_backend_avalible']
                ax_field.is_display_backend_avalible = the_item
            if ('is_columnn_avalible' in item and
                    ax_field.is_columnn_avalible !=
                    item['is_columnn_avalible']):
                ax_field.is_columnn_avalible = item['is_columnn_avalible']
            if ('is_updated_always' in item
                    and ax_field.is_updated_always != item['is_updated_always']):
                ax_field.is_updated_always = item['is_updated_always']
            if 'is_group' in item and ax_field.is_group != item['is_group']:
                ax_field.is_group = item['is_group']
            if ('is_always_whole_row' in item and
                    ax_field.is_always_whole_row !=
                    item['is_always_whole_row']):
                ax_field.is_always_whole_row = item['is_always_whole_row']


def collect_ax_stats():
    """ Collect Ax usage statistics """
    with ax_model.scoped_session() as db_session:
        values = {}
        values['host'] = os.environ.get('AX_HOST', None)
        values['url'] = os.environ.get('AX_URL', None)
        values['dialect'] = os.environ.get('AX_DB_DIALECT', None)
        values['cache'] = os.environ.get('AX_CACHE_MODE', None)
        values['workersNum'] = os.environ.get('AX_SANIC_WORKERS', None)
        values['fingerprint'] = uuid.getnode()
        values['clientGuid'] = ax_auth.client_guid
        values['axVersion'] = os.environ.get('AX_VERSION', None)
        values['usersNum'] = db_session.query(AxUser).filter(
            AxUser.is_group.is_(False)).count()
        values['formsNum'] = db_session.query(AxForm).filter(
            AxForm.is_folder.is_(False)).count()
        # dbSize - #TODO how to do it?
        # sslUsed
        if os.environ.get('SSL_CERT_ABSOLUTE_PATH', None):
            values['sslUsed'] = True

        # actionCodeUsed
        if db_session.query(AxAction).filter(AxAction.code.isnot(None)).first():
            values['actionCodeUsed'] = True

        # gridCodeUsed
        if db_session.query(AxGrid).filter(AxGrid.code.isnot(None)).first():
            values['gridCodeUsed'] = True

        # roleCodeUsed
        if db_session.query(AxRole).filter(AxRole.code.isnot(None)).first():
            values['roleCodeUsed'] = True

        # adminEmails
        admin_emails = []
        admin_group = db_session.query(AxUser).filter(
            AxUser.is_admin.is_(True)
        ).filter(
            AxUser.is_group.is_(True)
        ).first()
        if admin_group:
            for usr in admin_group.users:
                if usr.email != 'default@ax-workflow.com':
                    admin_emails.append(usr.email)

        values['adminEmails'] = ", ".join(admin_emails)

        # fieldUsage
        # ax_fields = db_session.query(AxField.field_type_tag).filter(
        #     AxField.is_tab.is_(False)
        # ).all()
        # field_dict = {}
        # for fld in ax_fields:
        #     if fld.field_type_tag not in field_dict:
        #         field_dict[fld.field_type_tag] = 1
        #     else:
        #         field_dict[fld.field_type_tag] += 1
        # values['fieldUsage'] = field_dict
        this.stats = values


async def send_stats():
    """ Sends usage statistics to Ax sidekick server """
    # data = {
    #     "host": "hostHost",
    #     "usersNum": 10
    # }
    value_str = json.dumps(this.stats).replace('"', '\\"')
    query_str = (
        "    mutation{"
        "        doAction("
        "            formDbName: \"AxUsageStats\""
        "            actionDbName: \"newStats\""
        f"            values: \"{value_str}\""
        "        ) {"
        "            ok"
        "        }"
        "    }"
    )
    json_data = {'query': query_str}
    ret_str = await ax_misc.post_json(
        'http://ax-workflow.com/ax_stats', json_data, 5)
    return ret_str


def sync_field_types():
    """ Read field_types.yaml and check if any fields need to be
        created or modified """
    field_types_yaml = ax_misc.path('field_types.yaml')

    with ax_model.scoped_session() as db_session:

        ax_field_types = db_session.query(AxFieldType).all()
        existing_tags = []
        for field_type in ax_field_types:
            existing_tags.append(field_type.tag)

        if not os.path.isfile(field_types_yaml):
            raise FileNotFoundError(
                'Configuration failed, field_types.yaml not found')

        with open(field_types_yaml, 'r') as stream:
            yaml_vars = yaml.safe_load(stream)
            for item in yaml_vars:
                if isinstance(item, dict) and 'tag' in item:
                    tag = item['tag']
                    if tag in existing_tags:
                        update_field_type(
                            field_types=ax_field_types, item=item)
                    else:
                        create_field_type(db_session=db_session, item=item)

            db_session.commit()


# def database_fits_metadata() -> None:
#     """Compare metada and database"""
#     try:
#         with ax_model.engine.connect() as active_connection:
#             context = MigrationContext.configure(active_connection)
#             db_session = ax_model.Session()
#             ax_model.Base.query = db_session.query_property()
#             metadata = ax_model.Base.metadata
#             diff = compare_metadata(context, metadata)
#             return diff == []
#     except Exception:
#         logger.exception('Error in comparing metadata and database.')
#         raise


def database_fits_metadata() -> None:
    """Compare yaml db version and version in actual database"""
    with ax_model.scoped_session() as db_session:
        alembic_version = db_session.query(AxAlembicVersion).first()
        version = alembic_version.version_num
        yaml_version = os.environ.get('AX_DB_REVISION')
        if yaml_version == version:
            return True
        return False


def upgrade_database():
    """Run Alembic migration script and update db version"""
    try:
        version = os.environ.get('AX_DB_REVISION')
        logger.info(f'WARNING! Upgrading database to version {version}')
        command.upgrade(this.alembic_cfg, "head")
        command.stamp(this.alembic_cfg, "head")
    except Exception:
        logger.exception('Error upgrading database with Alembic')
        raise


def init_migration():
    """Initiate migration module"""
    init_alembic_config()
    if tables_exist() is False:
        create_tables()
        sync_field_types()
        create_default_users()
        create_default_pages()
    else:
        if database_fits_metadata() is False:
            upgrade_database()

        sync_field_types()
        collect_ax_stats()
        try:
            asyncio.get_event_loop().run_until_complete(send_stats())
        except Exception:
            pass
